import os
from distutils import dir_util
import re
import subprocess
import shutil
import codecs
import datetime
import tempfile
from contextlib import contextmanager
import hashlib


class Path(object):
    @property
    def directory(self):
        return Dirpath(os.path.dirname(self.path))

    def __init__(self, *bits):
        self.path = ''
        if bits:
            bits = list(bits)
            bits[0] = self.normalize(bits[0])
            for i in range(1, len(bits)):
                bits[i] = bits[i].strip('\\/')
            self.path = os.path.join(*bits)

    @classmethod
    def normalize(cls, d):
        """completely normalize a relative path (a path with ../, ./, or ~/)"""
        return os.path.abspath(os.path.expanduser(str(d)))

    @classmethod
    def get_temp(cls, *bits):
        tmp = tempfile.gettempdir()
        return cls(tmp, *bits)

    @classmethod
    def create_temp(cls, *bits):
        instance = cls.get_temp(*bits)
        instance.create()
        return instance

    def __str__(self):
        return self.path

    def __unicode__(self):
        return unicode(self.__str__())

    def chmod(self, permissions):
        permissions = int(permissions)
        if permissions < 100 or permissions > 777:
            raise ValueError("permissions should be between 100 and 777")
        subprocess.check_call([
            "chmod",
            "{0:04}".format(permissions),
            self.path
        ])

    def chown(self, user):
        subprocess.check_call([
            "chown",
            user,
            self.path
        ])

    def chgrp(self, user):
        subprocess.check_call([
            "chgrp",
            user,
            self.path
        ])

    def move(self, path):
        p = type(self)(path)
        shutil.move(self.path, p.path)
        self.path = p.path

    def rename(self, name):
        p = type(self)(self.directory, name)
        return self.move(p)


class Dirpath(Path):
    def count(self, regex=None):
        count = 0
        regexp = re.compile(regex, re.I) if regex else None
        for root_dir, directories, files in os.walk(self.path):
            if regexp:
                count += len([f for f in files if regexp.search(f)])
                count += len([d for d in directories if regexp.search(d)])

            else:
                count += len(files) + len(directories)

        return count

    def __len__(self):
        return self.count()

    def create(self):
        """create the directory path"""
        return dir_util.mkpath(self.path)

    def exists(self):
        return os.path.isdir(self.path)

    def clear(self):
        """this will clear a directory path of all files and folders"""
        # http://stackoverflow.com/a/1073382/5006
        for root, dirs, files in os.walk(self.path, topdown=True):
            for td in dirs:
                shutil.rmtree(os.path.join(root, td))

            for tf in files:
                os.unlink(os.path.join(root, tf))

            break

    def delete(self):
        try:
            shutil.rmtree(self.path)

        except OSError:
            pass

    def files(self, regex=None):
        regexp = re.compile(regex, re.I) if regex else None
        for root_dir, _, files in os.walk(self.path, topdown=True):
            for basename in files:
                if regexp and not regexp.search(basename):
                    continue

                yield Filepath(os.path.join(root_dir, basename))

            break

    def delete_files(self, regex):
        """delete all files matching regex"""
        if not regex:
            raise ValueError("regex was empty")

        count = 0
        for f in self.files(regex):
            count += 1
            f.delete()

        return count

    def create_file(self, name, contents=""):
        """create the file with basename in this directory with contents"""
        output_file = Filepath(self.path, name)
        oldmask = os.umask(0)
        with output_file.open('w+') as f:
            f.truncate(0)
            f.seek(0)
            f.write(contents)
        oldmask = os.umask(oldmask)
        return output_file


class Filepath(Path):
    @property
    def checksum(self):
        """return -- string -- the md5 hash of the file"""
        h = hashlib.md5()
        # http://stackoverflow.com/a/21565932/5006
        blocksize = 204800
        with self.open() as fp:
            for block in iter(lambda: fp.read(blocksize), ""):
                h.update(block)
        return h.hexdigest()

    @property
    def name(self):
        return os.path.basename(self.path)

    @property
    def modified(self):
        # http://stackoverflow.com/a/1526089/5006
        t = os.path.getmtime(self.path)
        return datetime.datetime.fromtimestamp(t)

    def __iter__(self):
        with open(self.path, "r") as f:
            for line in f:
                yield line

    def contents(self):
        with open(self.path, "r") as f:
            return f.read()

    def lines(self):
        """this is different than python built-in lines() method in that it strips
        the line endings from the end of the string"""
        with open(self.path, "r") as f:
            for line in f:
                yield line.rstrip()

    def lc(self):
        """return line count"""
        return len(list(self.lines()))

    def exists(self):
        return os.path.isfile(self.path)

    def clear(self):
        self.write("")

    def delete(self):
        try:
            os.unlink(self.path)

        except OSError:
            pass

    def copy(self, dest_path):
        """copy this file to dest_path"""
        dest_path = self.normalize(dest_path)
        if os.path.isdir(dest_path):
            basename = self.name
            dest_file = type(self)(dest_path, basename)

        else:
            dest_file = type(self)(dest_path)

        return shutil.copy(self.path, dest_file.path)

    def backup(self, suffix=".bak", ignore_existing=True):
        """backup the file to the same directory with given suffix

        suffix -- str -- what will be appended to the file name (eg, foo.ext becomes
            foo.ext.bak)
        ignore_existing -- boolean -- if True overright an existing backup, if false
            then don't backup if a backup file already exists
        return -- instance -- an instance of Filepath with the backup filepath
        """
        path = "{}{}".format(self.path, suffix)
        bak = type(self)(path)
        if ignore_existing or not bak.exists():
            self.copy(path)
        return bak

    def append(self, contents):
        with self.open("a") as f:
        #with codecs.open(self.path, encoding='utf-8', mode='a') as f:
            ret = f.write(contents)

        return ret

    def open(self, mode="r"):
        return codecs.open(self.path, encoding='utf-8', mode=mode)

    def write(self, contents, mode="w+"):
        with self.open(mode) as f:
        #with codecs.open(self.path, encoding='utf-8', mode='w+') as f:
            #f.truncate(0)
            #f.seek(0)
            ret = f.write(contents)

        return ret

    def writelines(self, lines, mode="w+"):
        """this is different than built-in python writelines in that it adds the
        newlines at the end of each line"""
        with self.open(mode) as f:
            ret = f.writelines(("{}\n".format(line) for line in lines))

        return ret

    def create(self):
        """create the file if it doesn't exist, touch the file if it does"""
        # http://stackoverflow.com/a/1160227/5006

        with open(self.path, 'a') as f:
            os.utime(self.path, None)

        return

        #with open(self.path, 'a') as f:
        with open(self.path, 'w+') as f:
            os.utime(self.path, None)
#             f.flush()
#             os.fsync(f.fileno())

    def contains(self, regex, flags=0):
        flags |= re.M
        m = None
        try:
            m = re.search(regex, self.contents(), flags=flags)

        except IOError:
            # TODO -- check for file not existing EOINT or whatever and only
            # suppress that
            # raised when file doesn't exist
            pass
        return True if m else False

    def delete_lines(self, regex, flags=0):
        """returns how many lines were removed"""
        count = 0
        regexp = re.compile(regex, flags)
        with self.open("r") as f:
            lines = f.readlines()

        with self.open("w") as f:
            f.truncate(0)
            for line in lines:
                if regexp.search(line):
                    count += 1
                else:
                    f.write(line)

        return count

    def modified_within(self, seconds=0, **timedelta_kwargs):
        """returns true if the file has been modified within the last seconds

        seconds -- integer -- how many seconds
        **timedelta_kwargs -- dict -- can be thing like days, must be passed in
            named
        return -- boolean -- True if file has been modified after the seconds back
        """
        if not self.exists(): return False

        now = datetime.datetime.now()
        then = self.modified
        timedelta_kwargs["seconds"] = seconds
        td_check = datetime.timedelta(**timedelta_kwargs)
        #pout.v(now, td_check, then, (now - td_check))
        return (now - td_check) < then


class Sentinal(object):
    def __init__(self, name):
        self.name = name
        now = datetime.datetime.utcnow()
        self.year = now.strftime("%Y")
        self.month = now.strftime("%m")
        self.f = Filepath.get_temp("{}-{}-{}".format(self.name, self.year, self.month))

    def exists(self):
        ret = False
        try:
            ret = self._exists

        except AttributeError:
            ret = self.f.exists()

        finally:
            self._exists = ret

        return ret

    def create(self):
        self.f.create()
        self._exists = True

    @classmethod
    @contextmanager
    def check(cls, name):
        """yields a boolean, execute to check if the block whould be run"""
        instance = cls(name)
        exists = instance.exists()
        yield instance
        if not exists:
            instance.create()

    def __str__(self):
        return self.f.path

    def __nonzero__(self):
        return self.exists()




from unittest import TestCase, SkipTest

import testdata

from stockton import dns
from stockton.dns import Domain
from stockton.path import Filepath, Dirpath


class DomainTest(TestCase):
    def test_mx(self):
        d = Domain("example.com")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "{} mail is handled by 10 mailstore1.secureserver.net.".format(d.host),
                "{} mail is handled by 0 smtp.secureserver.net.".format(d.host),
                "",
            ])
        testdata.patch(d, query=monkey_query)

        mxs = d.mx()
        self.assertEqual(2, len(mxs))

        mxs = d.mx("smtp.secureserver.net")
        self.assertEqual(1, len(mxs))

        vs = d.nameservers("000.00.000.000")
        self.assertEqual(0, len(vs))

    def test_nameservers(self):
        d = Domain("example.com")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "{} name server ns62.domaincontrol.com.".format(d.host),
                "{} name server ns61.domaincontrol.com.".format(d.host),
                ""
            ])
        testdata.patch(d, query=monkey_query)

        vs = d.nameservers()
        self.assertEqual(2, len(vs))

        vs = d.nameservers("ns62.domaincontrol.com")
        self.assertEqual(1, len(vs))

        vs = d.nameservers("000.00.000.000")
        self.assertEqual(0, len(vs))


    def test_a(self):
        d = Domain("example.com")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "{} has address 100.70.100.200".format(d.host),
                "{} has address 101.71.101.201".format(d.host),
                ""
            ])
        testdata.patch(d, query=monkey_query)

        vs = d.a()
        self.assertEqual(2, len(vs))

        vs = d.a("101.71.101.201")
        self.assertEqual(1, len(vs))

        vs = d.a("000.00.000.000")
        self.assertEqual(0, len(vs))

    def test_txt(self):
        d = Domain("example.com")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "{} descriptive text \"google-site-verification=dakfkdkaureurjkfdhiqerukSHA\"".format(d.host),
                "{} descriptive text \"v=spf1 include:_spf.google.com include:sendgrid.net include:spf.mandrillapp.com ~all\"".format(d.host),
                "{} descriptive text \"v=spf1 include:servers.mcsv.net ?all\"".format(d.host),
                ""
            ])
        testdata.patch(d, query=monkey_query)

        vs = d.txt()
        self.assertEqual(3, len(vs))

        vs = d.txt("spf")
        self.assertEqual(2, len(vs))

        vs = d.txt("000.00.000.000")
        self.assertEqual(0, len(vs))

    def test_ptr(self):
        d = Domain("123.456.789.100")
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "203.68.131.104.in-addr.arpa domain name pointer {}.".format("example.com"),
                ""
            ])
        testdata.patch(d, query=monkey_query)

        vs = d.ptr()
        self.assertEqual(1, len(vs))

        vs = d.ptr("example.com")
        self.assertEqual(1, len(vs))

        vs = d.ptr("000.00.000.000")
        self.assertEqual(0, len(vs))

    def test_rdns(self):
        def monkey_query(self, *args, **kwargs):
            return "\n".join([
                "123.456.789.100.in-addr.arpa domain name pointer {}.".format(d.host),
                ""
            ])
        MonkeyDomain = testdata.patch(
            Domain,
            a=lambda *a, **kw: ["123.456.789.100"],
            query=monkey_query
        )

        d = MonkeyDomain("example.com")
        vs = d.rdns()
        self.assertEqual("123.456.789.100", vs[0])

#     def test_domainkey(self):
#         import sys
#         import os
#         sys.path.insert(0, os.curdir)
# 
#         monkey_text_valid = "".join([
#             'v=DKIM1; k=rsa; ',
#             'p=MIIBIjANBgkqhkiG9wcheQEFAAOCAQ8AMIIBCgKCAQEAy14JM1EVS+y5CsPo',
#             'xvokcnYKlHVownd7A8RqcW0Ndb/PMZy1htsLgskDLwVbwp4TjxaNi6Wxakt0Kz',
#             'xeLT6AC7vmg0zHyMUzy0ra6sWyg3lPfool6wHKlF0ary2KQmbd6yyN+AyiQIT6',
#             'kq+E7hqyElnuAWUjA/Irnwlr2aZTBkQ3jUOY4c9IPa2FkYYdBuRASCAL8d0',
#             'rvoCqPpF+xvQF4W1uVjNx14wUcm739LAW+1Uw6VATrxZDp7QRhJd35zDQdwena',
#             'BbWVelqWm2RoTE0BARU6mTHsD3bO1OvwwqBS9uw/scLxpPW0AwlSkksUOSzKI2',
#             'FMNNyn0kKGk2ZJFWUowIDAQAB'
#         ])
#         monkey_text_from_host = "".join([
#             'v=DKIM1\; k=rsa\; ',
#             'p=MIIBIjANBgkqhkiG9wcheQEFAAOCAQ8AMIIBCgKCAQEAy14JM1EVS+y5CsPo',
#             'xvokcnYKlHVownd7A8RqcW0Ndb/PMZy1htsLgskDLwVbwp4TjxaNi6Wxakt0Kz',
#             'xeLT6AC7vmg0zHyMUzy0ra6sWyg3lPfool6wHKlF0ary2KQmbd6yyN+AyiQIT6',
#             'kq+E7hqyElnuAWUjA/Irnwlr2aZTBkQ3jUOY4c9IPa2FkYYdBuRAS" "CAL8d0',
#             'rvoCqPpF+xvQF4W1uVjNx14wUcm739LAW+1Uw6VATrxZDp7QRhJd35zDQdwena',
#             'BbWVelqWm2RoTE0BARU6mTHsD3bO1OvwwqBS9uw/scLxpPW0AwlSkksUOSzKI2',
#             'FMNNyn0kKGk2ZJFWUowIDAQAB'
#         ])
# 
#         class MonkeyDKIM(object):
#             def domainkey(*args, **kwargs):
#                 class OB(object): pass
#                 ob = OB()
#                 ob.subdomain = "foo._domainkey.example.com"
#                 ob.v = "DKIM1"
#                 ob.k = "rsa"
#                 ob.p = "foobar"
#                 ob.text = "RASCAL"
#                 return ob
# 
#         class MonkeyDomain(Domain):
#             def query(self, *args, **kwargs):
#                 return '{} descriptive text "{}"'.format(self.host, monkey_text_from_host)
# 
#         monkey_dns = testdata.patch(dns, Domain=MonkeyDomain, DKIM=MonkeyDKIM)
# 
#         d = monkey_dns.Domain("example.com")
#         self.assertTrue("RASCAL" in d.dkim()[0]["text"])
#         self.assertEqual(monkey_text_valid, d.dkim()[0]["text"])
#         return
# 
#         d = monkey_dns.Alias("example.com", "mail.example.com")
#         dkim = d.needed_dkim()
#         pout.v(dkim[1][1])
#         self.assertTrue("RASCAL" in dkim[1][1])

    def test_domainkey(self):
        d = Domain("example.com")

        monkey_text_valid = "".join([
            'v=DKIM1; k=rsa; ',
            'p=MIIBIjANBgkqhkiG9wcheQEFAAOCAQ8AMIIBCgKCAQEAy14JM1EVS+y5CsPo',
            'xvokcnYKlHVownd7A8RqcW0Ndb/PMZy1htsLgskDLwVbwp4TjxaNi6Wxakt0Kz',
            'xeLT6AC7vmg0zHyMUzy0ra6sWyg3lPfool6wHKlF0ary2KQmbd6yyN+AyiQIT6',
            'kq+E7hqyElnuAWUjA/Irnwlr2aZTBkQ3jUOY4c9IPa2FkYYdBuRASCAL8d0',
            'rvoCqPpF+xvQF4W1uVjNx14wUcm739LAW+1Uw6VATrxZDp7QRhJd35zDQdwena',
            'BbWVelqWm2RoTE0BARU6mTHsD3bO1OvwwqBS9uw/scLxpPW0AwlSkksUOSzKI2',
            'FMNNyn0kKGk2ZJFWUowIDAQAB'
        ])
        monkey_text_from_host = "".join([
            'v=DKIM1\; k=rsa\; ',
            'p=MIIBIjANBgkqhkiG9wcheQEFAAOCAQ8AMIIBCgKCAQEAy14JM1EVS+y5CsPo',
            'xvokcnYKlHVownd7A8RqcW0Ndb/PMZy1htsLgskDLwVbwp4TjxaNi6Wxakt0Kz',
            'xeLT6AC7vmg0zHyMUzy0ra6sWyg3lPfool6wHKlF0ary2KQmbd6yyN+AyiQIT6',
            'kq+E7hqyElnuAWUjA/Irnwlr2aZTBkQ3jUOY4c9IPa2FkYYdBuRAS" "CAL8d0',
            'rvoCqPpF+xvQF4W1uVjNx14wUcm739LAW+1Uw6VATrxZDp7QRhJd35zDQdwena',
            'BbWVelqWm2RoTE0BARU6mTHsD3bO1OvwwqBS9uw/scLxpPW0AwlSkksUOSzKI2',
            'FMNNyn0kKGk2ZJFWUowIDAQAB'
        ])
        def monkey_query(self, *args, **kwargs):
            return '{} descriptive text "{}"'.format(self.host, monkey_text_from_host)

        testdata.patch(d, query=monkey_query)

        self.assertTrue("RASCAL" in d.dkim()[0]["text"])
        self.assertEqual(monkey_text_valid, d.dkim()[0]["text"])

# class MailserverTest(TestCase):
#     def test_




import re
import time

from captain import echo

from .. import cli
from ..path import Filepath, Dirpath
from ..concur.formats.postfix import Main, SMTPd, Master
from ..concur.formats.generic import SpaceConfig
from ..geo import IP
from .base import Interface


class Cert(object):

    @property
    def key_f(self):
        return Filepath(self.certs_d, "{}.key".format(self.domain))

    @property
    def crt_f(self):
        return Filepath(self.certs_d, "{}.crt".format(self.domain))

    @property
    def pem_f(self):
        return Filepath(self.certs_d, "{}.pem".format(self.domain))

    def __init__(self, domain):
        self.domain = domain
        self.certs_d = Dirpath("/etc/postfix/certs")
        self.bits = 4096

    def assure(self):
        if not self.exists():
            self.create()

    def exists(self):
        return self.pem_f.exists()

    def create(self):
        """write out a certificate for the domain"""
        self.certs_d.create()
        domain = self.domain

        certs_key = self.key_f
        certs_crt = self.crt_f
        certs_pem = self.pem_f

        ip = IP()
        country = ip.country
        state = ip.state
        city = ip.city

        cli.package("openssl", only_upgrade=True)

        # http://superuser.com/questions/226192/openssl-without-prompt
        cli.run(" ".join([
            "openssl req",
            "-new",
            "-newkey rsa:{}".format(self.bits),
            # openssl 1.0.2+ only, comment out above line and uncomment next 2
            #"-newkey ec",
            #"-pkeyopt ec_paramgen_curve:prime256v1",
            "-days 3650",
            "-nodes",
            "-x509",
            "-subj \"/C={}/ST={}/L={}/O={}/CN={}\"".format(
                country,
                state,
                city,
                domain,
                domain
            ),
            "-keyout {}".format(certs_key),
            "-out {}".format(certs_crt)
        ]))
        cli.run("cat {} {} > {}".format(certs_crt, certs_key, certs_pem))


class Postfix(Interface):
    @property
    def mailserver(self):
        m = self.main()
        return m["myhostname"].val

    @property
    def helo(self):
        return SpaceConfig(dest_path=self.helo_f.path)

    @property
    def helo_f(self):
        return Filepath(self.config_d, "helo.regexp")

    @property
    def main_f(self):
        return Filepath(Main.dest_path)

    @property
    def main_live(self):
        """return the main config prototyped with the live main.cf"""
        return self.main()

    @property
    def main_new(self):
        """return the main config with no prototype, so brand spanking new"""
        return self.main("")

    @property
    def master_f(self):
        return Filepath(Master.dest_path)

    @property
    def config_d(self):
        return Dirpath("/etc/postfix")

    @property
    def virtual_d(self):
        return Dirpath(self.config_d, "virtual")

    @property
    def addresses_d(self):
        return Dirpath(self.virtual_d, "addresses")

    @property
    def domains_f(self):
        return Filepath(self.virtual_d, "domains")

    @property
    def domains(self):
        """return all the domains postfix has configured"""
        try:
            domains = set(self.domains_f.lines())
        except IOError:
            domains = set()
        return domains

    @property
    def addresses(self):
        for domain in self.domains:
            yield self.address(domain)

    def cert(self, domain):
        return Cert(domain)

    def base_configs(self):
        return [self.main_f, self.master_f]

    def autodiscover_domain(self, proxy_f):
        """given a proxy file, discover the domain it represents by looking through
        each line and making sure the first email addresses all match"""
        if not proxy_f:
            raise ValueError("proxy_f does not contain a valid path")

        proxy_f = Filepath(proxy_f)
        domain_check = set()
        for line in proxy_f:
            m = re.match("^(\S*@\S+)", line)
            if m:
                bits = m.group(1).split("@", 1)
                if len(bits) == 2:
                    domain_check.add(bits[1])

        if len(domain_check) == 1:
            domain = domain_check.pop()

        else:
            raise ValueError("Found multiple domains in proxy_file {}".format(proxy_f))

        return domain

    def delete_domain(self, domain):
        # remove postfix settings
        virtual_d = self.virtual_d

        domains_f = self.domains_f
        domains_f.delete_lines("^{}$".format(domain))

        addresses_d = self.addresses_d
        addresses_d.delete_files("^{}".format(domain))

        domain_hashes = []
        for d in domains_f.lines():
            domain_f = Filepath(addresses_d, d)
            domain_hashes.append("hash:{}".format(domain_f.path))

        m = self.main()
        m["virtual_alias_maps"] = ",\n  ".join(domain_hashes)
        m.save()

    def add_domain(self, domain, proxy_file, proxy_email):
        if not proxy_file:
            if not domain or not proxy_email:
                raise ValueError("Either pass in proxy_file or (domain and proxy_emails)")

        # create directory /etc/postfix/virtual
        virtual_d = self.virtual_d
        virtual_d.create()

        # create addresses directory
        addresses_d = self.addresses_d
        addresses_d.create()

        domains_f = self.domains_f
        domains_f.create()
        old_domains = self.domains
        new_domains = set()

        if proxy_file:
            echo.h3("Adding domain {} with addresses from {} to postfix", domain, proxy_file)
            f = Filepath(proxy_file)
            domain_f = self.address(domain)
            f.copy(domain_f)
            new_domains.add(domain)

        elif proxy_email:
            echo.h3("Adding catchall for {} routing to {}", domain, proxy_email)
            domain_f = Filepath(addresses_d, domain)
            domain_c = SpaceConfig(dest_path=domain_f.path)
            domain_c["@{}".format(domain)] = proxy_email
            domain_c.save()
            new_domains.add(domain)

        domains_f.writelines(old_domains.union(new_domains))

        # this is an additive editing of the conf file, so we use the active conf
        # as the prototype
        m = self.main()
        m.update(
            ("virtual_alias_domains", domains_f.path),
            ("virtual_alias_maps", ",\n  ".join(("hash:{}".format(af) for af in self.addresses)))
        )
        m.save()

        for domain in new_domains:
            cli.run("postmap {}".format(self.address(domain)))

    def address(self, domain):
        addresses_f = Filepath(self.addresses_d, domain)
        return addresses_f

    def main(self, path=Main.dest_path):
        return Main(prototype_path=path)

    def master(self, path=Master.dest_path):
        path = str(path)
        return Master(prototype_path=path)

    def main_backups(self):
        """return any backups of the Postfix main.cf file"""
        for mbak_f in self.config_d.files("main.*?\.bak$"):
            mbak = Main(dest_path=mbak_f.path, prototype_path=mbak_f.path)
            yield mbak

    def _run(self, cmd):
        """this wraps the normal command in a command that will make sure it works"""
        #return cli.run('script -c "{}" -q STDOUT --return'.format(cmd))
        return cli.run('script -c "{}" -q /dev/null --return'.format(cmd))

    def start(self):
        try:
            self._run("postfix start")
        except RuntimeError as e:
            if not re.search("Postfix\s+mail\s+system\s+is\s+already\s+running", str(e), re.I):
                raise

    def restart(self):
        if self.is_running():
            self._run("postfix reload")
        else:
            self.start()

    def stop(self):
        o = self._run("postfix stop")
#         for x in range(20):
#             if self.is_running():
#                 time.sleep(0.1)
#             else:
#                 break

    def is_running(self):
        try:
            self._run("postfix status")
            ret = True

        except RuntimeError:
            ret = False

        return ret

    def exists(self):
        return self.config_d.exists()

    def install(self):
        cli.package("postfix")

    def uninstall(self):
        try:
            self.stop()
        except cli.RunError as e:
            if not e.is_missing():
                raise

        cli.purge("postfix")
        self.config_d.delete()




import re

from captain import echo

from .. import cli
from .postfix import Postfix
from ..path import Filepath, Dirpath
from ..concur.formats.opendkim import OpenDKIM
from .base import Interface


class DomainKey(object):

    bits = 2048

    @property
    def keys_d(self):
        dk = DKIM()
        return dk.keys_d

    @property
    def txt_f(self):
        txt_f = Filepath(self.keys_d, "{}.txt".format(self.domain))
        return txt_f

    @property
    def private_f(self):
        private_f = Filepath(self.keys_d, "{}.private".format(self.domain))
        return private_f

    @property
    def text(self):
        dkim_text = "{} {} {}".format(self.v, self.k, self.p)
        return dkim_text

    @property
    def subdomain(self):
        m = re.match("^(\S+)", self.contents())
        return "{}.{}".format(m.group(1), self.domain)

    @property
    def v(self):
        mv = re.search("v=\S+", self.contents())
        return mv.group(0)

    @property
    def k(self):
        mk = re.search("k=\S+", self.contents())
        return mk.group(0)

    @property
    def p(self):
        contents = self.contents()
        mp = re.search("p=[^\"]+", contents)
        mps = re.findall("\"(?!\S=)(\S+?)\"", contents)
        return "".join([mp.group(0)] + mps)

    def __init__(self, domain):
        self.domain = domain

    def exists(self):
        return self.txt_f.exists()

    def contents(self):
        try:
            contents = self._contents
        except AttributeError:
            contents = self._contents = self.txt_f.contents()
        return contents

    def create(self):
        if self.exists(): return
        self.generate()

    def generate(self):
        bits = self.bits
        domain = self.domain
        keys_d = self.keys_d
        private_f = self.private_f
        #cli.run("opendkim-genkey --domain={} --verbose --directory=\"{}\"".format(
        cli.run("opendkim-genkey --bits={} --domain={} --directory=\"{}\"".format(
            bits,
            domain,
            keys_d.path
        ))

        private_f = Filepath(keys_d, "default.private")
        private_f.rename("{}.private".format(domain))
        private_f.chmod(600)
        private_f.chown("opendkim:opendkim")

        txt_f = Filepath(keys_d, "default.txt")
        txt_f.rename("{}.txt".format(domain)) # this makes self.txt_f and self.exists() work

    def __str__(self):
        return self.text


class DKIM(Interface):

    bits = 2048

    @property
    def config_f(self):
        return Filepath(OpenDKIM.dest_path)

    @property
    def config_d(self):
        return Dirpath("/etc/opendkim")

    @property
    def keys_d(self):
        return Dirpath(self.config_d, "keys")

    @property
    def keytable_f(self):
        return Filepath(self.config_d, "KeyTable")

    @property
    def signingtable_f(self):
        return Filepath(self.config_d, "SigningTable")

    @property
    def trustedhosts_f(self):
        return Filepath(self.config_d, "TrustedHosts")

    def config(self, path=OpenDKIM.dest_path):
        return OpenDKIM(prototype_path=path)

    def base_configs(self):
        return [self.config_f]

    def domainkey(self, domain):
        domk = DomainKey(domain)
        domk.bits = self.bits
        return domk

    def set_ip(self, ip_addr):
        """set the ip address into the hosts file"""
        if not ip_addr:
            raise ValueError("ip_addr is empty")

        hosts = [
            "127.0.0.1",
            "::1",
            "localhost",
            "192.168.0.1/24",
            ip_addr,
        ]

        trustedhosts_f = self.trustedhosts_f
        trustedhosts_f.writelines(hosts)

    def delete_domain(self, domain):
        # remove dkim settings
        self.signingtable_f.delete_lines(domain)
        self.trustedhosts_f.delete_lines(domain)

    def add_domains(self):
        p = Postfix()
        for domain in p.domains:
            self.add_domain(domain)

    def add_domain(self, domain, gen_key=False):
        keytable_f = self.keytable_f
        signingtable_f = self.signingtable_f
        trustedhosts_f = self.trustedhosts_f

        dk = self.domainkey(domain)
        private_f = dk.private_f

        if gen_key:
            dk.generate()
        else:
            dk.create()

        if not keytable_f.contains(domain):
            keytable_f.append("{} {}:default:{}\n".format(
                dk.subdomain,
                domain,
                private_f.path
            ))

        if not signingtable_f.contains(domain):
            signingtable_f.append("{} {}\n".format(
                domain,
                dk.subdomain
            ))

        if not trustedhosts_f.contains(domain):
            trustedhosts_f.append("*.{}\n".format(domain))

    def start(self):
        cli.run("/etc/init.d/opendkim start")

    def restart(self):
        if self.is_running():
            cli.run("/etc/init.d/opendkim restart")
        else:
            self.start()

    def stop(self):
        cli.run("/etc/init.d/opendkim stop")

    def is_running(self):
        ret = False
        try:
            output = cli.run("/etc/init.d/opendkim status")
            if re.search("opendkim\s+is\s+running", output, flags=re.I):
                ret = True

        except RuntimeError:
            ret = False

        return ret

    def exists(self):
        return self.config_d.exists()

    def install(self):
        cli.package("opendkim", "opendkim-tools")
        self.config_d.create()
        self.keys_d.create()

    def uninstall(self):
        cli.purge("opendkim", "opendkim-tools")
        self.config_d.delete()
        for f in self.base_configs():
            f.delete()




"""
Postfix specific configuration files, the file locations are for ubuntu 14.04
"""
#import re

from . import generic
from . import base


class SpamAssassinOption(base.ConfigOption):
    def format_set(self, name, divider, val):
        option_set_format = "{}{}{}" # key=val
        s = option_set_format.format(name, divider, val)
        return s


class SpamAssassin(generic.EqualConfig):
    option_class = SpamAssassinOption
    dest_path = "/etc/default/spamassassin"


class Local(generic.SpaceConfig):
    dest_path = "/etc/spamassassin/local.cf"




from unittest import TestCase as BaseTestCase, SkipTest
import os

from captain.client import Captain


class TestCase(BaseTestCase):
    @classmethod
    def setup_env(cls):
        pass

    @classmethod
    def setUpClass(cls):
        if os.environ["USER"] != "root":
            raise RuntimeError("User is not root, re-run this test with sudo")

        cls.setup_env()


class Stockton(Captain):
    """The command line runner for running one of the cli commands"""

    script_quiet = False

    def __init__(self, subcommand):
        self.cmd_prefix = "python -m stockton --verbose {}".format(subcommand.replace("_", "-"))
        super(Stockton, self).__init__("")




"""
Geo support 
"""
import tempfile
import urllib
import gzip
import os

import geoip2.database

from . import cli


class IP(str):
    @property
    def geo_ip(self):
        if not os.path.isfile(self.geo_path):
            self.download_geo()

        if not hasattr(self, "_geo_ip"):
            cls = type(self)
            if not hasattr(cls, "geo_db"):
                type(self).geo_db = geoip2.database.Reader(self.geo_path)
            self._geo_ip = cls.geo_db.city(self)

        return self._geo_ip

    @property
    def country(self):
        return self.geo_ip.country.iso_code

    @property
    def state(self):
        return self.geo_ip.subdivisions.most_specific.iso_code

    @property
    def city(self):
        return self.geo_ip.city.name

    def __new__(cls, ip_addr=""):
        if not ip_addr:
            ip_addr = cli.ip()
        instance = super(IP, cls).__new__(cls, ip_addr)
        instance.geo_path = os.path.join(tempfile.gettempdir(), "GeoLite2-City.mmdb")
        instance.geo_url = "http://geolite.maxmind.com/download/geoip/database/GeoLite2-City.mmdb.gz"
        return instance

    def download_geo(self):
        gz_path = "{}.gz".format(self.geo_path)
        # http://stackoverflow.com/a/31857152/5006
        urllib.urlretrieve(self.geo_url, gz_path)

        # TODO -- see if this works
        # https://docs.python.org/2/library/gzip.html
#         with open('file.txt', 'rb') as f_in, gzip.open('file.txt.gz', 'wb') as f_out:
#             shutil.copyfileobj(f_in, f_out)

        with gzip.open(gz_path, 'rb') as gzf:
            with open(self.geo_path, "w+b") as f:
                f.write(gzf.read())





import os
import time

import testdata

from . import TestCase as BaseTestCase
#from stockton.dns import Domain
#from stockton.path import Filepath, Dirpath
#from stockton.interface.dkim import DKIM
from stockton.interface import Postfix, DKIM, SRS, Spam
from stockton.path import Filepath, Dirpath


class TestCase(BaseTestCase):
    service_class = None

    @classmethod
    def create_instance(cls):
        return cls.service_class()

    @classmethod
    def setup_env(cls):
        s = cls.create_instance()
        s.install()

    def test_lifecycle(self):
        s = self.create_instance()
        s.start()
        self.assertTrue(s.is_running())

        s.stop()
        self.assertFalse(s.is_running())

        s.restart()
        self.assertTrue(s.is_running())


class PostfixTest(TestCase):
    service_class = Postfix

    def test_uninstall(self):
        p = Postfix()
        p.uninstall()
        self.assertFalse(p.config_d.exists())
        self.assertFalse(p.is_running())

    def test_stray_output_files(self):
        """make sure the typescript problem is no longer a problem"""
        # https://github.com/Jaymon/stockton/issues/26
        f = Filepath(os.getcwd(), "typescript")

        f.delete()
        self.assertFalse(f.exists())

        p = Postfix()
        self.assertTrue(p.is_running())

        self.assertFalse(f.exists())


class DKIMTest(TestCase):
    service_class = DKIM

    def test_uninstall(self):
        d = DKIM()
        d.uninstall()
        self.assertFalse(d.config_f.exists())
        self.assertFalse(d.config_d.exists())

    def test_txt_info(self):
        d = DKIM()
        d.bits = 8192
        domain = "txtinfo.com"
        d.add_domain(domain, gen_key=True)

        dk = d.domainkey(domain)
        self.assertEqual(d.bits, dk.bits)
        self.assertNotEqual(None, dk.v)
        self.assertNotEqual(None, dk.k)
        self.assertNotEqual(None, dk.p)

    def test_domainkey(self):
        """Turns out really long domain keys were having problems being parsed"""
        d = DKIM()

        domain = "foo.com"

        d.bits = 1024
        d.add_domain(domain, gen_key=True)
        dk = d.domainkey(domain)
        self.assertEqual(218, len(dk.p))

        d.bits = 2048
        d.add_domain(domain, gen_key=True)
        dk = d.domainkey(domain)
        self.assertEqual(394, len(dk.p))
        #pout.v(dk.p, dk.txt_f.contents())

        d.bits = 4096
        d.add_domain(domain, gen_key=True)
        dk = d.domainkey(domain)
        self.assertEqual(738, len(dk.p))

        d.bits = 8192
        d.add_domain(domain, gen_key=True)
        dk = d.domainkey(domain)
        self.assertEqual(1418, len(dk.p))


class SRSTest(TestCase):
    service_class = SRS


class SpamTest(TestCase):
    service_class = Spam

    def test_lifecycle(self):
        sp = Spam()
        c = sp.config()
        c.update(
            ("ENABLED", 1),
        )
        c.save()
        sp.start()

        super(SpamTest, self).test_lifecycle()

    def test_uninstall(self):
        sp = Spam()
        sp.uninstall()
        self.assertFalse(sp.is_running())
        self.assertFalse(sp.home_d.exists())


del TestCase



from ..path import Filepath, Dirpath

class Interface(object):
    def base_configs(self):
        return []

    def base_get(self, name, configs=None):
        if not configs:
            configs = self.base_configs()

        base_fs = []
        for f in configs:
            base_fs.append(Filepath("{}.{}.base".format(f.path, name)))

        return base_fs

    def base_create(self, name, configs=None):
        """create a snapshot of the relevant config files so we could restore if needed"""
        if not configs:
            configs = self.base_configs()

        base_fs = []
        for f in configs:
            if f.exists():
                base_f = f.backup(suffix=".{}.base".format(name), ignore_existing=False)
                base_fs.append(base_f)

        return base_fs

    def base_restore(self, name, configs=None):
        """restore a snapshot of the base"""
        if not configs:
            configs = self.base_configs()

        for f in configs:
            base_f = Filepath("{}.{}.base".format(f.path, name))
            if base_f.exists():
                base_f.copy(f)

    def base_clear(self, name=""):
        for f in self.base_configs():
            d = f.directory
            d.delete_files("{}.base$".format(name))

    def reset(self):
        """Set it back completely fresh, you should never use this unless you know
        what you're doing"""
        self.uninstall()
        self.install()

    def start(self):
        raise NotImplementedError()

    def restart(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def is_running(self):
        raise NotImplementedError()

    def install(self):
        raise NotImplementedError()

    def uninstall(self):
        raise NotImplementedError()




from unittest import TestCase

import testdata

from stockton.concur.formats.base import ConfigFile, Config, ConfigSection
from stockton.concur.formats import postfix
from stockton.concur.formats import generic


def setUpModule():
    pass


def tearDownModule():
    pass


# class ConfigFileTest(TestCase):
#     def test_replay(self):
#         path = testdata.create_file("counting", "\n".join([str(x) for x in range(5)]))
# 
#         class RConSec(ConfigSection):
#             def _parse(self, fp):
#                 self.name = fp.line
# 
#         class RCon(Config):
#             parse_class = RConSec
# 
# 
#         fp = ConfigFile(path, RCon())
# 
#         c = fp.next()
#         self.assertEqual(0, int(c.line))
# 
#         c = fp.next()
#         self.assertEqual(1, int(c.line))
# 
#         fp.replay(c, fp.line, fp.line_number)
#         c = fp.next()
#         self.assertEqual(1, int(c.line))
# 
#         c = fp.next()
#         self.assertEqual(2, int(c.line))
# 
#         c = fp.next()
#         self.assertEqual(3, int(c.line))
# 
#         fp.replay(c, fp.line, fp.line_number)
#         c = fp.next()
#         self.assertEqual(3, int(c.line))
# 
#         c = fp.next()
#         self.assertEqual(4, int(c.line))
# 
#         with self.assertRaises(StopIteration):
#             fp.next()


class SpaceTest(TestCase):
    def test__parse(self):
        path = testdata.create_file("space.conf", "\n".join([
            "# Log to syslog",
            "Syslog			yes",
            "# Required to use local socket with MTAs that access the socket as a non-",
            "# privileged user (e.g. Postfix)",
            "UMask			002",
            "",
            "# Sign for example.com with key in /etc/mail/dkim.key using",
            "# selector '2007' (e.g. 2007._domainkey.example.com)",
            "#Domain			example.com",
        ]))

        c = generic.SpaceConfig(prototype_path=path)
        self.assertTrue(c["syslog"])
        self.assertTrue(c["UMask"])
        self.assertTrue(c["Domain"])
        #self.assertEqual(3, len(c.options))


class PostfixTest(TestCase):
    def test_main_update(self):
        contents = "\n".join([
            "foo = bar",
            "che = baz",
        ])
        path = testdata.create_file("main.cf", contents)
        c = postfix.Main(dest_path=path, prototype_path=path)

        c.update(
            ("foo", "rab")
        )
        c.save()

        c2 = postfix.Main(dest_path=path, prototype_path=path)
        self.assertEqual("rab", c2["foo"].val)
        self.assertEqual("baz", c2["che"].val)

    def test_main_oneline(self):
        contents = "\n".join([
            "virtual_alias_map = hash:/some/path/one",
            "foo = bar",
            "che = baz",
        ])
        path = testdata.create_file("main.cf", contents)
        c = postfix.Main(prototype_path=path)
        self.assertEqual(contents, str(c))

    def test_main_multiline(self):
        contents = "\n".join([
            "virtual_alias_map = hash:/some/path/one,",
            "  hash:/some/path/two,",
            "  hash:/some/path/three",
        ])
        path = testdata.create_file("main.cf", contents)
        c = postfix.Main(prototype_path=path)
        self.assertEqual(contents, str(c))

        contents = "\n".join([
            "virtual_alias_map = hash:/some/path/one,",
            "  hash:/some/path/two,",
            "  hash:/some/path/three",
            "virtual_alias_domains = /another/path/and/stuff",
        ])
        path = testdata.create_file("main.cf", contents)
        c = postfix.Main(prototype_path=path)
        self.assertEqual(contents, str(c))

        c["virtual_alias_map"] = c["virtual_alias_map"].val + "\n  hash:/some/path/four"
        contents = "\n".join([
            "virtual_alias_map = hash:/some/path/one,",
            "  hash:/some/path/two,",
            "  hash:/some/path/three",
            "  hash:/some/path/four",
            "virtual_alias_domains = /another/path/and/stuff",
        ])
        self.assertEqual(contents, str(c))

    def test_master_multiple_same_name(self):
        contents = "\n".join([
            "smtp      inet  n       -       -       -       -       smtpd",
            "#smtp      inet  n       -       -       -       1       postscreen",
            "smtp      unix  -       -       -       -       -       smtp",
        ])
        master_path = testdata.create_file("master.cf", contents)

        master = postfix.Master(prototype_path=master_path)
        for smtp in master["smtp"]:
            smtp.chroot = "n"

        contents = "\n".join([
            "smtp      inet  n       -       n       -       -       smtpd",
            "smtp      inet  n       -       n       -       1       postscreen",
            "smtp      unix  -       -       n       -       -       smtp",
        ])
        self.assertEqual(contents, str(master))

    def test_master_option(self):
        master = postfix.Master()
        class FP(object):
            line = "  -o foo=yes"

        mo = master.create_option()
        mo.parse(FP())
        self.assertEqual("foo", mo.name)
        self.assertEqual("yes", mo.val)

        mo.val = "no"
        self.assertEqual("no", mo.val)
        self.assertEqual("  -o foo=no", str(mo))


    def test_master_section_manipulate_option(self):
        master_path = testdata.create_file("master.cf", "\n".join([
            "smtp      inet  n       -       -       -       -       smtpd",
            "  -o foo=yes",
            "  -o bar=yes",
        ]))

        master = postfix.Master(prototype_path=master_path)
        self.assertTrue(isinstance(master["smtp"], postfix.MasterSection))

        contents = "\n".join([
            "smtp      inet  n       -       -       -       -       smtpd",
            "  -o foo=no",
            "  -o bar=yes",
        ])
        master["smtp"]["foo"] = "no"
        self.assertEqual(contents, str(master))

    def test_master_section_append_option(self):
        master_path = testdata.create_file("master.cf", "\n".join([
            "smtp inet n - - - - smtpd",
            "  -o foo=yes",
        ]))

        master = postfix.Master(prototype_path=master_path)
        contents = "\n".join([
            "smtp inet n - - - - smtpd",
            "  -o foo=yes",
            "  -o bar=yes",
        ])
        master["smtp"]["bar"] = "yes"
        self.assertEqual(contents, str(master))

    def test_master_sections(self):
        contents = "\n".join([
            "one inet n - - - - smtpd",
            "  -o smtpd_tls_wrappermode=yes",
            "two inet n - - - - qmqpd",
            "three fifo n - - 60 1 pickup",
        ])
        master_path = testdata.create_file("master.cf", contents)

        master = postfix.Master(prototype_path=master_path)
        self.assertEqual(contents, str(master))

    def test_master_section_commented(self):
        contents = "#smtps     inet  n       -       -       -       -       smtpd"
        master_path = testdata.create_file("master.cf", contents)

        master = postfix.Master(prototype_path=master_path)
        master["smtps"].modified = True
        self.assertEqual("smtps     inet  n       -       -       -       -       smtpd", str(master))

    def test_master_comments(self):
        master_path = testdata.create_file("master.cf", "\n".join([
            "#",
            "# this is the stuff before the first section",
            "#",
            "# ==========================================================================",
            "# service type  private unpriv  chroot  wakeup  maxproc command + args",
            "#               (yes)   (yes)   (yes)   (never) (100)",
            "# ==========================================================================",
            "smtp      inet  n       -       -       -       -       smtpd",
            "  -o smtpd_tls_wrappermode=yes",
            "  -o smtpd_sasl_auth_enable=yes",
            "  -o smtpd_client_restrictions=permit_sasl_authenticated,reject",
            "  -o milter_macro_daemon_name=ORIGINATING",
        ]))

        master = postfix.Master(prototype_path=master_path)
        self.assertEqual(8, len(master.lines))
        self.assertEqual(4, len(master["smtp"].lines))

    def test_master_section_modify_option(self):
        master_path = testdata.create_file("master.cf", "\n".join([
            "#submission inet n       -       -       -       -       smtpd",
            "#  -o syslog_name=postfix/submission",
            "#  -o smtpd_tls_security_level=encrypt",
            "#  -o smtpd_sasl_auth_enable=yes",
            "#  -o smtpd_reject_unlisted_recipient=no",
            "#  -o smtpd_client_restrictions=one",
            "#  -o smtpd_helo_restrictions=two",
            "#  -o smtpd_sender_restrictions=three",
            "#  -o smtpd_recipient_restrictions=",
            "#  -o smtpd_relay_restrictions=permit_sasl_authenticated,reject",
            "#  -o milter_macro_daemon_name=ORIGINATING"
        ]))

        m = postfix.Master(prototype_path=master_path)
        m["submission"].modified = True
        m["submission"].update(
            ("syslog_name", "postfix/submission"),
            ("smtpd_tls_security_level", "may"),
            ("smtpd_tls_cert_file", "/foo/bar.pem"),
            ("smtpd_sasl_auth_enable", "yes"),
            ("smtpd_reject_unlisted_recipient", "no"),
            ("smtpd_relay_restrictions", "permit_sasl_authenticated,reject"),
            ("milter_macro_daemon_name", "ORIGINATING")
        )
        self.assertEqual(4, str(m).count("#"))
        self.assertEqual(11, len(m["submission"].options))

    def test_master_add_section(self):
#         master_path = testdata.create_file("master.cf", "\n".join([
#             "smtp      inet  n       -       -       -       -       smtpd",
#         ]))
        master = postfix.Master(prototype_path="")
        section = master.create_section("smtp inet n - - - - smtpd")
        self.assertEqual(0, len(master.sections))
        master["smtp"] = section
        self.assertEqual(1, len(master.sections))
        self.assertEqual("smtp", master["smtp"].name)



class ConfigTest(TestCase):
    def test_update_before(self):
        contents = "\n".join([
            "foo = 1",
            "baz = 3",
        ])
        path = testdata.create_file("update_before.conf", contents)
        conf = Config(prototype_path=path)

        conf.update_before("baz", ("bar", 2))
        self.assertTrue("foo" in str(conf.lines[0]))
        self.assertTrue("bar" in str(conf.lines[1]))
        self.assertTrue("baz" in str(conf.lines[2]))

    def test_options(self):
        contents = "\n".join([
            "foo=bar",
            "baz = che",
        ])
        path = testdata.create_file("main.cf", contents)

        conf = Config(prototype_path=path)
        self.assertEqual(contents, str(conf))

        contents = "\n".join([
            "foo=bar",
            "baz = che",
        ])
        conf["foo"] = "che"
        conf["baz"] = "bar"

        contents = "\n".join([
            "foo = che",
            "baz = bar",
        ])
        self.assertEqual(contents, str(conf))




"""
Postfix specific configuration files, the file locations are for ubuntu 14.04
"""
import re

from . import generic
from . import base


class MainOption(base.ConfigOption):
    def _parse(self, fp):
        super(MainOption, self)._parse(fp)
        if self.is_valid():
            # let's make sure we don't have a multiline
            line_number = fp.line_number
            for c in fp:
                m = re.match("^\s+(\S+)", c.line)
                if m:
                    line = "\n" + fp.line
                    self.val += line
                    self.line += line
                    line_number += 1

                else:
                    #self.lines = re.split(",?\s*", self.val, re.M)
                    fp.rewind(line_number)
                    break


class Main(generic.EqualConfig):
    option_class = MainOption
    dest_path = "/etc/postfix/main.cf"
    # /usr/share/postfix/main.cf.dist has a rather complete example file


class SMTPd(generic.ColonConfig):
    dest_path = "/etc/postfix/sasl/smtpd.conf"


class MasterSection(base.ConfigSection):
    def _parse(self, fp):
        if not re.search("\s+-\s+", fp.line): return

        self.reset()
        commenters = self.config.commenters

        # service type  private unpriv  chroot  wakeup  maxproc command + args",
        m = re.match("^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)", fp.line)
        if m:
            self.line = fp.line
            self.start_line_number = fp.line_number
            self.stop_line_number = self.start_line_number

            self.name = m.group(1).lstrip(commenters).lstrip()
            self.service_type = m.group(2)
            self.private = m.group(3)
            self.unpriv = m.group(4)
            self.chroot = m.group(5)
            self.wakeup = m.group(6)
            self.maxproc = m.group(7)
            self.cmd = m.group(8)

            for line_number, c in enumerate(fp):
                if isinstance(c, self.config.section_class):
                    self.stop_line_number = c.start_line_number - 1
                    fp.rewind(self.stop_line_number)
                    break

                elif isinstance(c, self.config.option_class):
                    self.options[c.name].append(line_number)
                    self.lines.append(c)

                elif isinstance(c, self.config.line_class):
                    self.lines.append(c)

    def __str__(self):
        if self.modified:
            s = "{}{}{}{}{}{}{}{}".format(
                self.name.ljust(max(10, len(self.name) + 1)),
                self.service_type.ljust(6),
                self.private.ljust(8),
                self.unpriv.ljust(8),
                self.chroot.ljust(8),
                self.wakeup.ljust(8),
                self.maxproc.ljust(8),
                self.cmd
            )

        else:
            s = self.line

        s2 = super(MasterSection, self).__str__()
        if s2:
            s += "\n" + s2
        return s


class MasterOption(base.ConfigOption):
    def _parse(self, fp):
        if "-o" in fp.line:
            self.name = fp.line

        commenters = self.config.commenters
        divider = self.config.option_divider
        m = re.match("^[{}]?\s*-o\s+(\S+)\s*{}\s*(.*)".format(commenters, divider), fp.line)
        if m:
            self.name = m.group(1)
            self.val = m.group(2)

    def format_set(self, name, divider, val):
        s = "  -o {}{}{}".format(name, divider, val)
        return s


class MasterLine(base.ConfigLine):
    def __str__(self):
        # TODO -- test this with an existing line to make sure it doesn't double indent
        return "  {}".format(self.line.lstrip())


class Master(base.Config):
    section_class = MasterSection
    option_class = MasterOption
    line_class = MasterLine
    dest_path = "/etc/postfix/master.cf"




"""
The base classes for all the other formats,

usually, you would actually use the classes in generics.py as the start for your custom
configuration parser, the classes in this module are the support classes for the
generics
"""
import re
from collections import defaultdict
import itertools


class ConfigBase(object):
    """the base common class for most of the other more useful classes"""
    @property
    def factory(self):
        try:
            cls = self.config
            cls.option_class
        except AttributeError:
            cls = self
            cls.option_class
        return cls

    def __init__(self, config):
        self.config = config
        self.reset()

    def reset(self):
        self.sections = defaultdict(list)
        self.options = defaultdict(list)

        # there are some things to keep in mind about lines, they don't have to be
        # true line numbers in the config file, they are line numbers in relation
        # to the content area you are in, so if the config file has 2 sections, then
        # there will only be 2 "lines" in that config file, one for each section
        self.lines = []

        self.name = ""
        self.val = ""

    def parse(self, fp):
        self.line = fp.line
        self._parse(fp)
        self.modified = False

    def is_valid(self):
        return True

    def create_option_for_key(self, k):
        """creates an option with name k, does not insert option into config file"""
        option = self.factory.create_option()
        option.name = k
        return option

    def update(self, *args, **kwargs):
        for body in itertools.chain(args, kwargs.items()):
            if isinstance(body, basestring):
                line = self.factory.create_line()
                line.line = body
                self.insert(len(self.lines), line)
            else:
                k, v = body
                self[k] = v

    def update_before(self, k, *args, **kwargs):
        """same as update but will insert all the args and kwargs before the option
        at k"""
        line_numbers = self.find_lines(k)
        if line_numbers:
            line_number = line_numbers[0] # the min line number you need to insert before
            for k, v in itertools.chain(args, kwargs.items()):
                if k in self:
                    self[k] = v

                else:
                    option = self.create_option_for_key(k)
                    option.val = v
                    self.insert(line_number, option)

        else:
            self.update(*args, **kwargs)

    def insert(self, line_number, option):
        """insert option before the given line_number, this is similar to the python
        built-in list.insert method, you need to use this function to keep all the
        internal data structurs in alignment"""
        if line_number < len(self.lines):
            # go through and move everything down to compensate for the
            # insert in the line_number list and line everything back up again
            for k in self.options.keys():
                k_line_numbers = self.options[k]
                for i in range(len(k_line_numbers)):
                    if k_line_numbers[i] > line_number:
                        k_line_numbers[i] += 1

            for k in self.sections.keys():
                k_line_numbers = self.sections[k]
                for i in range(len(k_line_numbers)):
                    if k_line_numbers[i] > line_number:
                        k_line_numbers[i] += 1

        self.lines.insert(line_number, option)
        if isinstance(option, self.factory.option_class):
            self.options[option.name].append(line_number)

        else:
            self.sections[option.name].append(line_number)

    def __setattr__(self, k, v):
        # http://stackoverflow.com/questions/17576009/python-class-property-use-setter-but-evade-getter
        if k == "modified":
            self.__dict__["modified"] = v

        else:
            self.modified = True
            super(ConfigBase, self).__setattr__(k, v)

    def __missing__(self, k):
        option = self.create_option_for_key(k)
        self.insert(len(self.lines), option)
        return option

    def __contains__(self, k):
        return k in self.sections or k in self.options

    def find_lines(self, k):
        """return how many lines the key is found in the config file"""
        line_numbers = []

        if k in self.sections:
            line_numbers = self.sections[k]

        elif k in self.options:
            line_numbers = self.options[k]

        return line_numbers

    def __getitem__(self, k):
        line_numbers = self.find_lines(k)

        # after we have the line count, we pull out that many lines, it's a string
        # if it is a one line value, or an array of lines if it encompassed more than
        # one line, if we didn't find the value at all, then create it
        if len(line_numbers) == 1:
            v = self.lines[line_numbers[0]]

        elif len(line_numbers) > 1:
            v = [self.lines[ln] for ln in line_numbers]

        else:
            v = self.__missing__(k)
            #raise KeyError("no section or option {}".format(k))

        return v

    def __setitem__(self, k, v):
        if isinstance(v, self.factory.section_class):
            if k in self.sections:
                self.sections[k] = v
            else:
                self.insert(len(self.lines), v)

        else:
            if k in self.sections:
                raise ValueError("you cannot modify sections using dict string notation")

            line_numbers = []

            if k in self.options:
                line_numbers = self.options[k]

            if len(line_numbers) > 0:
                for line_number in line_numbers:
                    option = self.lines[line_number]
                    option.val = v

            else:
                option = self.__missing__(k)
                option.val = v


class ConfigLine(ConfigBase):
    """This can be the base class for the different sections and options, or it
    can be a representation of a comment or some other line we don't usually do
    anything with"""
    def __str__(self):
        return self.line

    def _parse(self, fp):
        self.line = fp.line


class ConfigOption(ConfigLine):
    """This represents any key -> value pair in the configuration file"""
    def is_valid(self):
        return bool(self.name)

    def _parse(self, fp):
        line = fp.line
        name = ""
        val = ""
        comment = ""
        commenters = self.config.commenters
        divider = self.config.option_divider

        if re.match("^[^{}]\S+".format(commenters), line): # name = val
            name, val = re.split("\s*{}\s*".format(divider), line, 1)

        elif re.match("^[{}]\s*\S+\s*{}".format(commenters, divider), line): # # name = val
            l = re.sub("^[{}]\s*".format(commenters), "", line)
            name, val = re.split("\s*{}\s*".format(divider), l, 1)

        bits = re.split("\s[{}]\s*".format(commenters), val, 1)
        val = bits[0]
        if len(bits) > 1:
            comment = bits[1]

        self.name = name
        self.val = val
        self.comment = comment
        self.line = line
        self.modified = False

    def __str__(self):
        s = ""
        if self.modified:
            s = self.format_set(
                self.name,
                self.config.option_divider,
                self.val
            )
            comment = getattr(self, "comment", "")
            if comment:
                s += " {} {}".format(self.commenters[0], comment)

        else:
            s = super(ConfigOption, self).__str__()

        return s

    def format_set(self, name, divider, val):
        option_set_format = "{} {} {}" # key, divider, val
        s = option_set_format.format(name, divider, val)
        return s


class ConfigSection(ConfigOption):
    """certain configuration files might be broken up into sections of options, this
    handles representing those sections so when manipulating the config file you can
    add and rename sections, etc."""
    def _parse(self, fp):
        pass

    def __str__(self):
        s = "\n".join(str(cl) for cl in self.lines)
        return s


class ConfigFile(object):
    """This handles parsing the file internally"""
    def __init__(self, path, config):
        self.path = path
        self.config = config
        # we load the whole config file into memory because we hate memory and to
        # punish it we like to use lots of it. Also, we were having some difficulty
        # getting some conf files to parse correctly without being able to seek and
        # stuff to certain lines and replay parts of the file and this was the 
        # easiest solution to that problem
        with open(self.path, "r") as fp:
            self.lines = fp.readlines()
            self.count = len(self.lines)

        self.line_number = -1

    def close(self):
        if not self.fp.closed:
            self.fp.close()

    def __iter__(self):
        return self

    def __count__(self):
        return self.count

    def rewind(self, line_number):
        self.line_number = line_number

    def next(self):
        self.line_number += 1
        if self.line_number >= self.count:
            raise StopIteration()

        self.line = self.lines[self.line_number].rstrip()

        c = self.config.create_section()
        c.parse(self)
        if not c.is_valid():
            c = self.config.create_option()
            c.parse(self)
            if not c.is_valid():
                c = self.config.create_line()
                c.parse(self)

        return c


class ConfigBody(ConfigFile):
    def __init__(self, body, config):
        self.config = config
        self.lines = body.splitlines(False)
        self.count = len(self.lines)
        self.line_number = -1

    def close(self): pass


class Config(ConfigBase):
    """config parser that allows you to modify the parsed file and write it back out

    we don't use shlex because it didn't preserve comments
    https://docs.python.org/2/library/shlex.html 

    If an ini specific parser is added at a later date, use this?
    https://docs.python.org/2/library/configparser.html

    You can modify the config file in 3 different ways:

        1) item notation, config["option"] = "value"

            using the item notation, you set the value, either a string or list
            of strings

        2) update() notations, config.update([key, val], [key2, val2], ...)

            you send tuples of key, value pairs to the update method, the reason
            you pass in tuples is so order is preserved

        3) get item, for advanced manipulation of the option

            this is the only way to manipulate a section

            opt = config["option"]
            opt.val = "value"
    """

    commenters = "#"

    option_divider = "="

    line_class = ConfigLine

    option_class = ConfigOption

    section_class = ConfigSection

    file_class = ConfigFile

    body_class = ConfigBody

    dest_path = ""

    prototype_path = ""

    def __init__(self, dest_path="", prototype_path=""):
        self.reset()

        if prototype_path:
            self.prototype_path = prototype_path
            self.parse()

        if dest_path:
            self.dest_path = dest_path

    def create_line(self):
        return self.line_class(self)

    def create_option(self):
        return self.option_class(self)

    def create_section(self, body=""):
        if body:
            config_s = self.body_class(body, self)
            for sc in config_s:
                break
            sc.modified = True

        else:
            sc = self.section_class(self)

        return sc

    def create_file(self):
        return self.file_class(self.prototype_path, self)

    def parse(self):
        fp = self.create_file()
        self.reset()

        for line_number, c in enumerate(fp):
            if isinstance(c, self.section_class):
                self.sections[c.name].append(line_number)
                self.lines.append(c)

            elif isinstance(c, self.option_class):
                self.options[c.name].append(line_number)
                self.lines.append(c)
 
            elif isinstance(c, self.line_class):
                self.lines.append(c)

    def __str__(self):
        return "\n".join((str(cl) for cl in self))

    def __iter__(self):
        for cl in self.lines:
            yield cl

    def save(self):
        with open(self.dest_path, "w") as f:
            for cl in self:
                f.write(str(cl))
                f.write("\n")




import os
import sys
import subprocess
import re
import hashlib

from captain import echo

from .path import Filepath


class RunError(RuntimeError):
    # http://tldp.org/LDP/abs/html/exitcodes.html
    def __init__(self, cmd, code, output, e=None):
        self.cmd = cmd
        self.code = code
        self.output = output
        self.e = e
        message = "{} returned {} with output {}".format(
            cmd,
            code,
            output
        )
        super(RunError, self).__init__(message)

    def is_missing(self):
        return self.code == 127

    def is_general(self):
        return self.code == 1

    def is_misuse(self):
        return self.code == 2

    def is_permissions(self):
        return self.code == 126

    def is_bad_arg(self):
        return self.code == 128

    def is_terminated(self):
        return self.code == 130

    def search(self, regex):
        ret = False
        if re.search(regex, str(self), re.I):
            ret = True
        return ret


def ip():
    #external_ip = cached_run("wget -qO- http://ifconfig.me/ip", ttl=10800)
    external_ip = cached_run("wget -qO- http://icanhazip.com", ttl=10800)
    external_ip = external_ip.strip()
    m = re.match("^(?:\d+\.){3}\d+$", external_ip)
    return external_ip if m else ""


def cached_run(cmd, ttl=3600, **process_kwargs):
    # http://stackoverflow.com/questions/5297448/how-to-get-md5-sum-of-a-string
    cmd_h = hashlib.md5(cmd).hexdigest()
    cmd_h = "{}-{}".format(os.environ["USER"], cmd_h)
    f = Filepath.get_temp(cmd_h)
    if f.modified_within(seconds=ttl):
        output = f.contents()

    else:
        process_kwargs["capture_output"] = True
        output = run(cmd, **process_kwargs)
        f.write(output)

    return output


def run(cmd, capture_output=False, **process_kwargs):
    # we will allow overriding of these values
    echo.out("< {}", cmd)

    output = ""
    process_kwargs.setdefault("stderr", subprocess.STDOUT)

    # we will not allow these to be overridden via kwargs
    process_kwargs["shell"] = True
    process_kwargs["stdout"] = subprocess.PIPE
    #process_kwargs["cwd"] = self.cwd

    try:
        process = subprocess.Popen(
            cmd,
            **process_kwargs
        )

        for line in iter(process.stdout.readline, ""):
            output += line
            echo.verbose(line)

        line = process.stdout.read()
        process.wait()
        if process.returncode > 0:
            raise RunError(cmd, process.returncode, output)

    except subprocess.CalledProcessError as e:
        raise RunError(cmd, e.returncode, e.output, e)

    return output


def package(*packages, **kwargs):

    # http://serverfault.com/questions/143968/automate-the-installation-of-postfix-on-ubuntu
    os.environ["DEBIAN_FRONTEND"] = "noninteractive"

    for p in packages:
        if kwargs.get("only_upgrade", False):
            run("apt-get -y install --only-upgrade {}".format(p))

        else:
            run("apt-get -y install --no-install-recommends {}".format(p))


def purge(*packages):

    os.environ["DEBIAN_FRONTEND"] = "noninteractive"

    for p in packages:
        #run("apt-get purge -y {}".format(p))
        run("apt-get -y remove --purge --auto-remove {}".format(p))




from unittest import TestCase
import time

import testdata

from stockton.path import Dirpath, Filepath, Sentinal


def setUpModule():
    pass


def tearDownModule():
    pass


class SentinalTest(TestCase):
    def test_exists(self):
        s = Sentinal(testdata.get_ascii())
        self.assertFalse(s.exists())

        s.create()
        self.assertTrue(s.exists())

    def test_with(self):
        name = testdata.get_ascii()
        with Sentinal.check(name) as execute:
            self.assertFalse(execute)

        s = Sentinal(name)
        self.assertTrue(s)


class FilepathTest(TestCase):
    def test_create(self):
        d = testdata.create_dir()
        fex = testdata.create_file("create_exists", "foo", d)
        f = Filepath(fex)
        self.assertTrue(f.exists())
        f.create()
        self.assertTrue("foo" in f.lines())

        f = Filepath(d, "create_not_exists")
        self.assertFalse(f.exists())
        f.create()
        self.assertTrue(f.exists())
        self.assertFalse("foo" in f.lines())

    def test_checksum(self):
        f = Filepath.create_temp(testdata.get_uuid())
        checksum = f.checksum
        f.write("foo")
        checksum2 = f.checksum
        self.assertNotEqual(checksum, checksum2)

        f.write("bar")
        time.sleep(0.1)
        checksum3 = f.checksum
        self.assertNotEqual(checksum, checksum3)
        self.assertNotEqual(checksum2, checksum3)

    def test_delete_lines(self):
        f = Filepath.create_temp("delete_lines")
        f.writelines([
            "foo",
            "foo",
            "bar",
            "foo"
        ])

        count = f.delete_lines("bar")
        self.assertEqual(1, count)
        self.assertEqual(3, len(list(f.lines())))
        self.assertFalse("bar" in f.contents())

        f = Filepath.create_temp("delete_lines_todelete")
        f.writelines([
            "todelete.com default._domainkey.todelete.com",
            ""
        ])
        count = f.delete_lines("todelete.com")
        self.assertEqual(1, count)
        self.assertFalse(f.contains("todelete.com"))


    def test_modified_within(self):
        f = Filepath.create_temp(testdata.get_ascii(32))
        self.assertTrue(f.modified_within(5))
        self.assertTrue(f.modified_within(days=1))
        time.sleep(1.1)
        self.assertFalse(f.modified_within(1))




import os
import re

import testdata

from stockton import cli
from stockton.path import Filepath, Dirpath
from stockton.concur.formats.postfix import Main, SMTPd, Master
from stockton.interface import Postfix, Spam, DKIM, SRS, SMTP

from . import TestCase as BaseTestCase, Stockton


class TestCase(BaseTestCase):
#     def setUp(self):
#         super(TestCase, self).setUp()

    @classmethod
    def setup_env(cls):
        p = Postfix()
        if not p.main_f.exists():
            s = Stockton("prepare")
            r = s.run()

    def setup_domain(self, domain):
        s = Stockton("configure-recv")
        r = s.run("--mailserver=mail.example.com")

        s = Stockton("add-domain")
        r = s.run("{} --proxy-email=dest@dest.com".format(domain))


class PrepareTest(TestCase):
    @classmethod
    def setup_env(cls):
        cli.purge("postfix")

    def test_run(self):
        p = Postfix()
        d = p.config_d
        self.assertFalse(d.exists())

        s = Stockton("prepare")
        r = s.run()

        f = p.main_f
        self.assertTrue(f.exists())


class ConfigureRecvTest(TestCase):
    @classmethod
    def setup_env(cls):
        super(ConfigureRecvTest, cls).setup_env()

        s = Stockton("configure-recv")
        r = s.run("--mailserver=mail.example.com")

    def test_recv(self):
        p = Postfix()

        # make some changes to main
        m = p.main()
        m.update(
            ("myhostname", "foobar.com")
        )
        m.save()

        f = p.main_f
        self.assertTrue(f.contains("foobar.com"))

        # re-run
        s = Stockton("configure-recv")
        r = s.run("--mailserver=mail.example.com")
        f = p.main_f
        self.assertFalse(f.contains("foobar.com"))
        self.assertTrue(p.is_running())


class ConfigureSendTest(TestCase):
    def test_send(self):
        p = Postfix()
        s = Stockton("configure-send")
        r = s.run("--mailserver=mail.example.com")
        self.assertTrue(p.is_running())


class ConfigureDKIMTest(TestCase):
    @classmethod
    def setup_env(cls):
        dk = DKIM()
        dk.uninstall()
        super(ConfigureDKIMTest, cls).setup_env()

    def test_dkim(self):
        #self.test_recv() # we need a configured for receive postfix
        self.setup_domain("example.com")

        s = Stockton("configure-dkim")
        r = s.run()

        dk = DKIM()
        p = Postfix()
        self.assertTrue(dk.config_d.exists())
        self.assertEqual(1, dk.keytable_f.lc())
        self.assertEqual(1, dk.signingtable_f.lc())
        self.assertTrue("*.example.com" in dk.trustedhosts_f.lines())
        self.assertTrue(dk.is_running())
        self.assertTrue(p.is_running())


class ConfigureSRSTest(TestCase):
    def test_srs(self):
        s = Stockton("configure-srs")
        r = s.run("")

        sr = SRS()
        self.assertTrue(sr.is_running())


class DomainTest(TestCase):
#     @classmethod
#     def setup_env(cls):
#         dk = DKIM()
#         dk.uninstall()
#         super(DomainTest, cls).setup_env()

    def test_delete_domain(self):
        self.setup_domain("example.com")
        s = Stockton("configure_dkim")
        r = s.run()

        s = Stockton("add-domain")
        r = s.run("todelete.com --proxy-email=todelete@dest.com")

        s = Stockton("delete-domain")
        r = s.run("todelete.com")

        # verify
        af = Filepath("/etc/postfix/virtual/addresses/todelete.com")
        self.assertFalse(af.exists())

        df = Filepath("/etc/postfix/virtual/domains")
        self.assertFalse(df.contains("todelete.com"))

        opendkim_d = Dirpath("/etc/opendkim")
        signingtable_f = Filepath(opendkim_d, "SigningTable")
        self.assertFalse(signingtable_f.contains("todelete.com"))

        trustedhosts_f = Filepath(opendkim_d, "TrustedHosts")
        self.assertFalse(trustedhosts_f.contains("todelete.com"))

    def test_add_domain_proxy_file_smtp(self):
        self.setup_domain("example.com")
        s = Stockton("configure-send")
        r = s.run("--mailserver=mail.example.com")

        proxy_addresses = testdata.create_file("foo.com", [
            "one@foo.com                foo@dest1.com",
            "two@foo.com                foo@dest2.com",
            "three@foo.com              foo@dest3.com",
            "",
        ])

        s = Stockton("add-domain")
        r = s.run("foo.com --proxy-file={} --smtp-password=12345".format(proxy_addresses))

        cli.package("cyrus-clients-2.4") # for smtptest
        r = cli.run(
            "echo QUIT | smtptest -a smtp@foo.com -w 12345 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost",
            capture_output=True
        )
        self.assertRegexpMatches(r, "235[^A]+Authentication\s+successful")

    def test_update_domain_proxy(self):
        # to make sure this works completely, we completely remove postfix
        p = Postfix()
        p.reset()

        f = testdata.create_file("foo.com", [
            "one@foo.com                foo@dest.com",
            "",
        ])

        s = Stockton("add-domain")
        r = s.run("{} --proxy-file={}".format(f.name, f))
        m = Main(prototype_path=Main.dest_path)
        proxy_path = re.search("^hash:([/a-z]+?foo\.com)$", m["virtual_alias_maps"].val, re.M).group(1)
        proxy_path_f = Filepath(proxy_path)
        self.assertTrue("one@foo.com" in proxy_path_f.contents())

        s = Stockton("update-domain-proxy")
        f = testdata.create_file("foo.com", [
            "two@bar.com                bar@dest.com",
            "",
        ])
        r = s.run("{} --proxy-file={}".format(f.name, f))
        m = Main(prototype_path=Main.dest_path)
        proxy_path = re.search("^hash:([/a-z]+?foo\.com)$", m["virtual_alias_maps"].val, re.M).group(1)
        proxy_path_f = Filepath(proxy_path)
        self.assertFalse("one@foo.com" in proxy_path_f.contents())
        self.assertTrue("two@bar.com" in proxy_path_f.contents())

    def test_add_domain_proxy_file_no_smtp(self):

        # to make sure this works completely, we completely remove postfix
        p = Postfix()
        p.reset()

        s = Stockton("add-domain")
        proxy_domains = testdata.create_files({
            "foo.com": "\n".join([
                "one@foo.com                foo@dest.com",
                "two@foo.com                foo@dest.com",
                "three@foo.com              foo@dest.com",
                "",
            ]),
            "bar.com": "\n".join([
                "one@bar.com                bar@dest.com",
                "two@bar.com                bar@dest.com",
                "three@bar.com              bar@dest.com",
                "",
            ]),
        })

        for f in proxy_domains:
            r = s.run("{} --proxy-file={}".format(f.name, f))

        contents = "\n".join([
            "@foo.com                   foo@dest.com",
            "",
        ])
        f = testdata.create_file("foo.com", contents)

        r = s.run("{} --proxy-file={}".format(f.name, f))
        foo = Filepath("/etc/postfix/virtual/addresses/foo.com")
        self.assertEqual(contents, foo.contents())

        # Let's check structure because I was having a lot of problems with getting
        # the structure...

        # domains file should have 2 domains in it
        domains_f = Filepath("/etc/postfix/virtual/domains")
        self.assertEqual(2, domains_f.lc())
        self.assertTrue(domains_f.contains("^foo\.com$"))

        m = Main(prototype_path=Main.dest_path)
        # we can't guarrantee foo, bar order so we match one line at a time
        self.assertTrue(re.search("^hash:[/a-z]+?(bar|foo)\.com,", m["virtual_alias_maps"].val, re.M))
        self.assertTrue(re.search("^\s+hash:[/a-z]+?(bar|foo)\.com$", m["virtual_alias_maps"].val, re.M))

    def test_add_domains(self):
        s = Stockton("add-domains")

        # a file with different domains
        proxy_domains = testdata.create_files({
            "foo2.com.txt": [
                "one@foo3.com                foo@dest.com",
                "one@bar3.com                bar@dest.com",
                "",
            ],
        })
        with self.assertRaises(RuntimeError):
            r = s.run("--proxy-domains={}".format(proxy_domains))

        # 2 files with the same domain
        proxy_domains = testdata.create_files({
            "foo2.com.txt": [
                "one@foo2.com                foo@dest.com",
                "",
            ],
            "bar2.com": [
                "one@foo2.com                bar@dest.com",
                "",
            ],
        })
        with self.assertRaises(RuntimeError):
            r = s.run("--proxy-domains={}".format(proxy_domains))

        proxy_domains = testdata.create_files({
            "foo.com.txt": [
                "one@foo.com                foo@dest.com",
                "two@foo.com                foo@dest.com",
                "three@foo.com              foo@dest.com",
                "",
            ],
            "bar.com": [
                "one@bar.com                bar@dest.com",
                "two@bar.com                bar@dest.com",
                "three@bar.com              bar@dest.com",
                "",
            ],
            "che.org.txt": [
                "@che.org                   che@dest.com",
                "",
            ],
        })

        r = s.run("--proxy-domains={} --smtp-password=1234".format(proxy_domains))
        self.assertTrue("Adding domain foo.com" in r)
        self.assertTrue("Adding domain bar.com" in r)
        self.assertTrue("Adding domain che.org" in r)

    def test_add_domain_domain(self):
        """pass in the domain and the proxy"""
        s = Stockton("configure-dkim")
        r = s.run()

        p = Postfix()
        dk = DKIM()

        s = Stockton("add-domain")
        one_f = p.address("one.com")
        two_f = p.address("two.com")
        domains_f = p.domains_f

        r = s.run("one.com --proxy-email=one@dest.com")
        self.assertTrue(one_f.contains("^@one.com"))
        self.assertTrue(domains_f.contains("one.com"))

        r = s.run("two.com --proxy-email=two@dest.com")
        self.assertTrue(one_f.contains("^@one.com"))
        self.assertTrue(domains_f.contains("one.com"))
        self.assertTrue(two_f.contains("two@dest.com"))
        self.assertTrue(domains_f.contains("two.com"))

        opendkim_d = dk.config_d
        keytable_f = dk.keytable_f
        signingtable_f = dk.signingtable_f
        trustedhosts_f = dk.trustedhosts_f

        self.assertTrue(keytable_f.contains("one.com"))
        self.assertTrue(signingtable_f.contains("one.com"))
        self.assertTrue(keytable_f.contains("two.com"))
        self.assertTrue(signingtable_f.contains("two.com"))
        self.assertTrue("*.one.com" in trustedhosts_f.lines())
        self.assertTrue("*.two.com" in trustedhosts_f.lines())

    def test_add_domain_smtp(self):
        s = Stockton("configure-send")
        r = s.run("--mailserver=mail.example.com")

        s = Stockton("add-domain")
        r = s.run("one.com --proxy-email=one@dest.com --smtp-password=1234")

        cli.package("cyrus-clients-2.4") # for smtptest
        r = cli.run(
            "echo QUIT | smtptest -a smtp@one.com -w 1234 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost"
        )
        self.assertRegexpMatches(r, "235[^A]+Authentication\s+successful")

        r = cli.run(
            "echo QUIT | smtptest -a smtp@one.com -w 9876 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost"
        )
        self.assertRegexpMatches(r, "535[^E]+Error:\s+authentication\s+failed")

        r = cli.run(
            "echo QUIT | smtptest -a foo@one.com -w 1234 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost"
        )
        self.assertRegexpMatches(r, "535[^E]+Error:\s+authentication\s+failed")




#         cli.package("cyrus-clients-2.4") # for smtptest
#
#         s = Stockton("configure-send")
#         arg_str = "example.com --mailserver=mail.example.com --smtp-password=1234"
#         r = s.run(arg_str)
# 
#         r = cli.run(
#             "echo QUIT | smtptest -a smtp@example.com -w 1234 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost",
#             capture_output=True
#         )
#         self.assertRegexpMatches(r, "235[^A]+Authentication\s+successful")
# 
#         r = cli.run(
#             "echo QUIT | smtptest -a smtp@example.com -w 9876 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost",
#             capture_output=True
#         )
#         self.assertRegexpMatches(r, "535[^E]+Error:\s+authentication\s+failed")
# 
#         r = cli.run(
#             "echo QUIT | smtptest -a foo@example.com -w 1234 -t /etc/postfix/certs/mail.example.com.pem -p 587 localhost",
#             capture_output=True
#         )
#         self.assertRegexpMatches(r, "535[^E]+Error:\s+authentication\s+failed")
# 
# #         r = cli.run(
# #             "echo QUIT | smtptest -a smtp -w 1234 -t /etc/postfix/certs/example.com.pem -p 587 localhost",
# #             capture_output=True
# #         )
# #         self.assertRegexpMatches(r, "535[^E]+Error:\s+authentication\s+failed")


class LockdownTest(TestCase):
    def test_all(self):
        s = Stockton("lockdown")
        r = s.run("--mailserver=mail.example.com")

        p = Postfix()
        sp = Spam()
        self.assertTrue(p.is_running())
        self.assertTrue(p.helo_f.exists())
        self.assertTrue(sp.is_running())

    def test_spam(self):
        s = Stockton("lockdown-spam")
        r = s.run()

        sp = Spam()
        self.assertTrue(sp.is_running())

    def test_postfix(self):
        self.setup_domain("example.com")

        s = Stockton("lockdown-postfix")
        r = s.run("--mailserver=mail.example.com")

        p = Postfix()
        self.assertTrue(p.is_running())
        self.assertTrue(p.helo_f.exists())


class InstallationTest(TestCase):
    @classmethod
    def setup_env(cls):
        #super(ConfigureDKIMTest, cls).setup_env()
        pass

    def test_idempotence(self):
        self.test_uninstall()
        self.test_install()

        p = Postfix()
        dk = DKIM()
        sp = Spam()
        sm = SMTP()

        main_hash = p.main_f.checksum
        master_hash = p.master_f.checksum
        dk_hash = dk.config_f.checksum
        sp_config_hash = sp.config_f.checksum
        sp_local_hash = sp.local_f.checksum
        sm_config_hash = sm.config_f.checksum

        self.test_install()

        # just to make sure there is no caching issues
        p = Postfix()
        dk = DKIM()
        sp = Spam()
        sm = SMTP()

        self.assertEqual(main_hash, p.main_f.checksum)
        self.assertEqual(master_hash, p.master_f.checksum)
        self.assertEqual(dk_hash, dk.config_f.checksum)
        self.assertEqual(sp_config_hash, sp.config_f.checksum)
        self.assertEqual(sp_local_hash, sp.local_f.checksum)
        self.assertEqual(sm_config_hash, sm.config_f.checksum)

    def test_uninstall(self):
        s = Stockton("uninstall")
        r = s.run()

        p = Postfix()
        self.assertFalse(p.is_running())
        self.assertFalse(p.exists())

        dk = DKIM()
        self.assertFalse(dk.is_running())
        self.assertFalse(dk.exists())

        sr = SRS()
        self.assertFalse(sr.is_running())

        sp = Spam()
        self.assertFalse(sp.is_running())
        self.assertFalse(sp.exists())

        sm = SMTP()
        self.assertFalse(sm.exists())

    def test_install_proxy_domains(self):
        s = Stockton("install")
        proxy_domains = testdata.create_files({
            "pd1.com.txt": [
                "one@pd1.com                foo@dest.com",
                "",
            ],
            "pd2.com": [
                "one@pd2.com                bar@dest.com",
                "",
            ],
        })

        r = s.run("--mailserver={} --proxy-domains={} --smtp-password=1234".format(
            "mail.example.com",
            proxy_domains
        ))
        # if there was no error than yay, it worked

    def test_install(self):
        s = Stockton("install")
        arg_str = " ".join([
            "--mailserver=mail.example.com",
            "--smtp-password=1234",
            "--proxy-email=d@dest.com",
            "front.com"
        ])
        r = s.run(arg_str)

        p = Postfix()
        self.assertTrue(p.is_running())

        dk = DKIM()
        self.assertTrue(dk.is_running())

        sr = SRS()
        self.assertTrue(sr.is_running())

        sp = Spam()
        self.assertTrue(sp.is_running())




#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import argparse
import os
import subprocess
import re

from captain import echo, exit, ArgError
from captain.decorators import arg, args

from stockton import __version__, cli, dns
from stockton.concur.formats.postfix import Main, SMTPd, Master
from stockton.concur.formats.opendkim import OpenDKIM
from stockton.concur.formats.generic import SpaceConfig
from stockton.path import Dirpath, Filepath, Sentinal

from stockton.interface import SMTP, Postfix, DKIM, Spam, SRS


def main_prepare():
    """Get a server ready to configure postfix by installing postfix"""
    echo.h2("Installing Postfix")

    with Sentinal.check("apt-get-update") as s:
        if not s:
            cli.run("apt-get update")

    #cli.run("apt-get -y install --no-install-recommends postfix")
    #cli.run("debconf-set-selections <<< \"postfix postfix/main_mailer_type string 'No configuration'\"")
    p = Postfix()
    p.install()


@arg('--mailserver', help='The domain mailserver (eg, mail.example.com)')
def main_configure_recv(mailserver):
    """Configure Postfix mailserver to receive email

    NOTE -- this doesn't actually configure any domains, in order to actually receive
    emails you should use add-domain(s)
    """
    # http://www.postfix.org/VIRTUAL_README.html

    echo.h2("Configuring Postfix to receive emails")

    p = Postfix()
    cert = p.cert(mailserver)
    cert.assure()

    settings = [
        ("alias_maps", "hash:/etc/aliases"), # http://unix.stackexchange.com/a/244200/118750
        ("myhostname", mailserver),
    ]

    # Smtpd means mails you receive from outside, smtp covers mails you send to other servers
    # http://blog.snapdragon.cc/2013/07/07/setting-postfix-to-encrypt-all-traffic-when-talking-to-other-mailservers/

    # Incoming
    settings.extend([
        ("smtpd_tls_cert_file", cert.crt_f.path),
        ("smtpd_tls_key_file", cert.key_f.path),
        ("smtpd_use_tls", "yes"),
        ("smtpd_tls_auth_only", "yes"),
        ("smtpd_tls_security_level", "may"),
        ("smtpd_tls_loglevel", 1),
        ("smtpd_tls_mandatory_ciphers", "high"),
        ("smtpd_tls_mandatory_protocols", "!SSLv2, !SSLv3"),
        ("smtpd_tls_session_cache_database", "btree:${data_directory}/smtpd_scache"),
    ])

    # Outgoing
    settings.extend([
        ("smtp_tls_cert_file", cert.crt_f.path),
        ("smtp_tls_key_file", cert.key_f.path),
        ("smtp_use_tls", "yes"),
        ("smtp_tls_security_level", "may"),
        ("smtp_tls_loglevel", 1),
        ("smtp_tls_mandatory_ciphers", "high"),
        ("smtp_tls_mandatory_protocols", "!SSLv2, !SSLv3"),
        ("smtp_tls_session_cache_database", "btree:${data_directory}/smtp_scache"),
    ])

    m = p.main()
    m.update(*settings)
    m.save()

    p.restart()


@arg('--mailserver', help='The domain mailserver (eg, mail.example.com)')
def main_configure_send(mailserver):
    """Configure postfix to handle smtp connections so you can send mail"""

    # https://help.ubuntu.com/lts/serverguide/postfix.html#postfix-sasl
    # http://www.postfix.org/SASL_README.html

    echo.h2("Configuring Postfix to send emails")

    p = Postfix()
    s = SMTP()

    s.install()

    c = s.config
    c.update(
        ("pwcheck_method", "auxprop"),
        ("auxprop_plugin", "sasldb"),
        ("mech_list", "PLAIN LOGIN CRAM-MD5 DIGEST-MD5 NTLM"),
        ("log_level", "7")
    )
    c.save()

    cert = p.cert(mailserver)
    cert.assure()

    #master_f = p.master_f
    #master_bak = p.base_create("send", [master_f])[0]

    # add the config to master.cf to enable smtp sending
    #m = p.master(master_bak.path)
    m = p.master()
    for smtp in m["smtp"]:
        if smtp.cmd == "smtpd":
            smtp.chroot = "n"

    m["submission"].chroot = "n"
    m["submission"].update(
        ("syslog_name", "postfix/submission"),
        ("smtpd_tls_security_level", "may"),
        ("smtpd_tls_cert_file", cert.pem_f.path),
        ("smtpd_sasl_auth_enable", "yes"),
        ("smtpd_reject_unlisted_recipient", "no"),
        ("smtpd_relay_restrictions", "permit_sasl_authenticated,reject"),
        ("milter_macro_daemon_name", "ORIGINATING")
    )
    m.save()

    p.restart()


def main_configure_dkim():
    """DomainKeys Identified Mail (DKIM)

    Setup dkim support

    http://www.dkim.org/
    """

    # setup for multi-domain support thanks to
    # http://askubuntu.com/questions/438756/using-dkim-in-my-server-for-multiple-domains-websites
    # http://seasonofcode.com/posts/setting-up-dkim-and-srs-in-postfix.html
    # http://edoceo.com/howto/opendkim
    # https://help.ubuntu.com/community/Postfix/DKIM

    echo.h2("Configuring Postfix to use DKIM")

    dk = DKIM()
    p = Postfix()

    dk.install()

    # could also use https://pypi.python.org/pypi/netifaces but this seemed easier
    # http://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
    external_ip = cli.ip()
    if external_ip:
        dk.set_ip(external_ip)

    trustedhosts_f = dk.trustedhosts_f
    c = dk.config()
    c.update(
        ("Canonicalization", "relaxed/simple"),
        ("Mode", "sv"),
        ("SubDomains", "yes"),
        ("Syslog", "yes"),
        ("LogWhy", "yes"),
        ("UMask", "022"),
        ("UserID", "opendkim:opendkim"),
        ("KeyTable", dk.keytable_f.path),
        ("SigningTable", dk.signingtable_f.path),
        ("ExternalIgnoreList", trustedhosts_f.path),
        ("InternalHosts", trustedhosts_f.path),
        ("Socket", "inet:8891@localhost")
    )
    c.save()

    m = p.main()
    m.update(
        ("milter_default_action", "accept"),
        # http://lists.opendkim.org/archive/opendkim/users/2011/08/1297.html
        ("milter_protocol", "6"),
        ("smtpd_milters", "inet:localhost:8891"),
        ("non_smtpd_milters", "inet:localhost:8891"),
    )
    m.save()

    # http://unix.stackexchange.com/a/74491/118750
    cli.run("adduser postfix opendkim")

    dk.add_domains()

    p.restart()
    dk.restart()


def main_configure_srs():
    """SRS: Sender Rewriting Scheme

    This configures SRS so SPF will work correctly and treat this MTA as a proxy

    http://www.openspf.org/SRS
    """

    echo.h2("Configuring Postfix to use SRS")

    s = SRS()
    p = Postfix()

    s.install()

    m = p.main()
    m.update(
        ("sender_canonical_maps", "tcp:localhost:10001"),
        ("sender_canonical_classes", "envelope_sender"),
        ("recipient_canonical_maps", "tcp:localhost:10002"),
        ("recipient_canonical_classes", "envelope_recipient,header_recipient")
    )
    m.save()

    s.restart()
    p.restart()


@arg('domain', help='The email domain to remove (eg, example.com)')
def main_delete_domain(domain):
    """remove a domain from the server"""
    echo.h3("Deleting domain {}", domain)

    # remove postfix settings
    p = Postfix()
    p.delete_domain(domain)

    # remove dkim settings
    dk = DKIM()
    dk.delete_domain(domain)

    p.restart()
    dk.restart()


@arg(
    '--proxy-domains',
    type=Dirpath,
    help='Directory containing domain configuration files'
)
@arg('--smtp-username', default="smtp", help='smtp username for sending emails')
@arg(
    '--smtp-password',
    default="",
    help='smtp password for sending emails, will be used for each domain in proxy-domains'
)
def main_add_domains(proxy_domains, smtp_username, smtp_password):
    """Given a directory containing domain alias configuration files, corresponding to
    the format specified in

    http://www.postfix.org/virtual.5.html

    Add each of the domains to Postfix, if smtp-password is given, then also set
    up the virtual domain to be able to send email."""
    p = Postfix()
    domains = {}
    for proxy_f in proxy_domains.files():
        domain = p.autodiscover_domain(proxy_f)
        if domain in domains:
            raise ValueError("proxy_file {} contains domain {} also seen in {}".format(
                proxy_f,
                domain,
                domains[domain],
            ))

        domains[domain] = proxy_f

    for domain, proxy_file in domains.items():
        echo.h2("Adding domain {} from file {}", domain, proxy_file)
        main_add_domain(domain, proxy_file, "", smtp_username, smtp_password)


@arg('domain', nargs="?", help='The email domain (eg, example.com)')
@arg('--proxy-file', default="", help='The file containing domain addresses to proxy emails')
@arg('--proxy-email', default="", help='The final destination email address')
def main_update_domain_proxy(domain, proxy_file, proxy_email):
    """don't do anything but update the virtual address file with the current in/out email
    addresses. If you need to do anything else, like set dkim or smtp passwords, use
    add-domain"""
    p = Postfix()

    if not domain:
        domain = p.autodiscover_domain(proxy_file)

    p.add_domain(domain, proxy_file, proxy_email)
    p.restart()


@arg('domain', nargs="?", help='The email domain (eg, example.com)')
@arg('--proxy-file', default="", help='The file containing domain addresses to proxy emails')
@arg('--proxy-email', default="", help='The final destination email address')
@arg('--smtp-username', default="smtp", help='smtp username for sending emails')
@arg('--smtp-password', default="", help='smtp password for sending emails')
def main_add_domain(domain, proxy_file, proxy_email, smtp_username, smtp_password):
    """add one virtual domain to postfix"""
    p = Postfix()

    if not domain:
        domain = p.autodiscover_domain(proxy_file)

    p.add_domain(domain, proxy_file, proxy_email)

    echo.h3("Configuring DKIM for {}", domain)
    dk = DKIM()
    if dk.exists():
        dk.add_domain(domain)
        dk.restart()

    if smtp_password:
        s = SMTP()
        s.add_user(smtp_username, smtp_password, domain)

    p.restart()
    main_check_domain(domain)


@arg('domain', help='The email domain (eg, example.com)')
def main_gen_domain_key(domain):
    """re-generate the DKIM key for the domain"""
    p = Postfix()
    dk = DKIM()
    dk.add_domain(domain, gen_key=True)
    main_check_domain(domain, ["dkim"])
    p.restart()
    dk.restart()


@arg('--mailserver', help='The domain mailserver (eg, mail.example.com)')
def main_lockdown_postfix(mailserver):
    """Tighten the postfix restrictions for sending and receiving email"""
    # http://www.cyberciti.biz/tips/postfix-spam-filtering-with-blacklists-howto.html
    # http://www.cyberciti.biz/faq/postfix-limit-incoming-or-receiving-email-rate/
    # https://www.debuntu.org/how-to-fight-spam-with-postfix-rbl/
    # https://www.howtoforge.com/virtual_postfix_antispam

    echo.h2("Locking down Postfix")

    p = Postfix()

    # make sure we've got a backup before we start messing with stuff, we won't
    # create a backup if it already exists
    main_f = p.main_f
    main_bak = main_f.backup(suffix=".lockdown.bak", ignore_existing=False)

    external_ip = cli.ip()

    echo.h3("configuring main")

    h = p.helo
    h.update(
        ("/^{}$/".format(re.escape(mailserver)), "550 Don't use my own hostname"),
        ("/^{}$/".format(re.escape(external_ip)), "550 Don't use my own IP address"),
        ("/^\[{}\]$/".format(re.escape(external_ip)), "550 Don't use my own IP address"),
        ("/^[0-9.]+$/", "550 Your software is not RFC 2821 compliant"),
        ("/^[0-9]+(\.[0-9]+){3}$/", "550 Your software is not RFC 2821 compliant")
    )
    h.save()

    m = p.main(main_bak.path)
    m.update(
        ("disable_vrfy_command", "yes"),
        ("smtpd_delay_reject", "yes"),
        ("strict_rfc821_envelopes", "yes"),
        ("smtpd_helo_required", "yes"),
        ("smtpd_helo_restrictions", ",\n    ".join([
            "permit_mynetworks",
            "permit_sasl_authenticated",
            "reject_non_fqdn_hostname",
            "reject_invalid_hostname",
            "regexp:{}".format(p.helo_f.path),
            "permit"
        ])),
        ("smtpd_recipient_restrictions", ",\n    ".join([
            "permit_mynetworks",
            "permit_sasl_authenticated",
            "reject_invalid_hostname",
            "reject_non_fqdn_hostname",
            "reject_non_fqdn_sender",
            "reject_non_fqdn_recipient",
            "reject_unknown_sender_domain",
            "reject_unknown_recipient_domain",
            "reject_unauth_destination",
            "reject_unknown_reverse_client_hostname",

            # from: http://www.akadia.com/services/postfix_spamassassin.html
            #reject_unauth_pipelining,
            #check_client_access hash:$config_directory/access_client,
            #check_sender_access hash:$config_directory/access_sender

            # disabled for spam assassin
            #"reject_rbl_client zen.spamhaus.org",
            #"reject_rbl_client bl.spamcop.net",
            #"reject_rbl_client b.barracudacentral.org",
            # disabled previously
            #"reject_rbl_client cbl.abuseat.org",
            #"reject_rbl_client dul.dnsbl.sorbs.net,",
            #"reject_rhsbl_sender dsn.rfc-ignorant.org",
            "permit"
        ])),

        # incoming
        ("smtpd_error_sleep_time", "1s"),
        ("smtpd_soft_error_limit", "10"),
        ("smtpd_hard_error_limit", "20"),

        # outgoing
        # This means that postfix will up to two concurrent
        # connections per receiving domains. The default value is 20.
        ("smtp_destination_concurrency_limit", 2),
        # Postfix will add a delay between each message to the same receiving domain.
        # It overrides the previous rule and in this example, it will send one email
        # after another with a delay of 1 second.
        ("smtp_destination_rate_delay", "5s"),
        # Limit the number of recipients of each message. If a message had 20 recipients
        # on the same domain, postfix will break it out to two different email messages instead of one.
        ("smtp_extra_recipient_limit", 10),
    )
    m.save()

    # dbl.spamhaus.org
    # xbl.spamhaus.org
    # b.barracudacentral.org
    # http://serverfault.com/a/514830/190381

    p.restart()


def main_lockdown_spam():
    """Install SpamAssassin and configure postfix to use it"""
    s = Spam()
    s.install()

    c = s.config()
    c.update(
        ("ENABLED", 1),
        ("OPTIONS", '"--create-prefs --max-children 5 --username {} --helper-home-dir {} -s /var/log/spamd.log"'.format(
            s.user,
            s.home_d
        )),
        ("CRON", 1)
    )
    c.save()

    c = s.local()
    c.update(
        ("rewrite_header", "Subject SPAM _SCORE_ *"),
        ("report_safe", 0),
        ("required_score", 5.0),
        ("use_bayes", 1),
        ("use_bayes_rules", 1),
        ("bayes_auto_learn", 1),
        ("skip_rbl_checks", 0),
        ("use_razor2", 0),
        #("use_dcc", 0), # SA didn't seem to like this line
        ("use_pyzor", 0),
    )
    c.save()

    # make backup of master.cf if it doesn't already exist
    p = Postfix()
#     master_bak = p.base_create("spam", [p.master_f])[0]
#     m = p.master(master_bak)
    m = p.master()

    # add the config to master.cf to enable smtp sending
    for smtp in m["smtp"]:
        if smtp.cmd == "smtpd":
            smtp.update(
                ("content_filter", "spamassassin")
            )

    section = m.create_section("spamassassin unix - n n - - pipe")
    section.update(
        "user={} argv=/usr/bin/spamc -f -e".format(s.user),
        "/usr/sbin/sendmail -oi -f ${sender} ${recipient}",
    )
    m["spamassassin"] = section
    m.save()

    p.restart()
    s.restart()


@args(main_lockdown_postfix, main_lockdown_spam)
def main_lockdown(mailserver):
    """Run all the lockdown commands"""
    main_lockdown_postfix(mailserver)
    main_lockdown_spam()


@arg('domain', help='The domain whose dns will be checked (eg, example.com)')
@arg(
    '--record', "-r",
    action="append",
    dest="records",
    help='records to check',
    default=None,
    choices=["mx", "a", "ptr", "spf", "dkim"],
)
def main_check_domain(domain, records=None):
    if not records:
        records = ["mx", "a", "ptr", "spf", "dkim"]

    pf = Postfix()
    mailserver = pf.mailserver
    external_ip = cli.ip()

    def print_record(name, domain, records):
        if records:
            echo.h1("DNS {} RECORD NEEDED FOR {}", name, domain)
            echo.br()

            echo.columns(*zip(*records))
            echo.br()

        else:
            echo.h3("{} record found", name)
            echo.br()

    if "a" in records:
        d = dns.Mailserver(mailserver, external_ip)
        print_record("A", mailserver, d.needed_a())

    if "ptr" in records:
        d = dns.Mailserver(mailserver, external_ip)
        print_record("PTR", mailserver, d.needed_ptr())

    if "mx" in records:
        d = dns.Alias(domain, mailserver)
        print_record("MX", domain, d.needed_mx())

    if "spf" in records:
        d = dns.Alias(domain, mailserver)
        print_record("SPF", domain, d.needed_spf())

    if "dkim" in records:
        try:
            d = dns.Alias(domain, mailserver)
            print_record("DKIM", domain, d.needed_dkim())
        except IOError:
            echo.h3("No local DKIM key found for {}", domain)


@args(main_configure_recv, main_add_domains, main_add_domain)
@arg('--proxy-domains', type=Dirpath, default=None)
def main_install(**kwargs):
    """Install, configure, and lockdown a postfix server"""
    main_prepare()
    main_configure_recv(mailserver=kwargs["mailserver"])
    main_configure_send(mailserver=kwargs["mailserver"])

    main_configure_dkim()
    main_configure_srs()
    main_lockdown(mailserver=kwargs["mailserver"])

    if kwargs["proxy_domains"]:
        main_add_domains(
            proxy_domains=kwargs["proxy_domains"],
            smtp_username=kwargs["smtp_username"],
            smtp_password=kwargs["smtp_password"]
        )

    else:
        main_add_domain(
            domain=kwargs["domain"],
            proxy_file=kwargs["proxy_file"],
            proxy_email=kwargs["proxy_email"],
            smtp_username=kwargs["smtp_username"],
            smtp_password=kwargs["smtp_password"]
        )


def main_uninstall():
    """Completely remove the postfix server from"""
    echo.h2("Uninstalling Mail Server")

    echo.h3("Removing SpamAssassin")
    sp = Spam()
    sp.uninstall()

    echo.h3("Removing SRS")
    sr = SRS()
    sr.uninstall()

    echo.h3("Removing DKIM")
    dk = DKIM()
    dk.uninstall()

    echo.h3("Removing SMTP")
    sm = SMTP()
    sm.uninstall()

    echo.h3("Removing Postfix")
    p = Postfix()
    p.uninstall()


def console():
    """we wrap captain.exit() so we can check for root"""
    if len(sys.argv) > 1 and "--help" not in sys.argv and "-h" not in sys.argv:
        if os.environ["USER"] != "root":
            raise RuntimeError("User is not root, re-run command with sudo")

    exit()

if __name__ == "__main__":
    console()




from unittest import TestCase
import time
import os

import testdata

from stockton import cli


class CliTest(TestCase):
    def test_stderr(self):
        filepath = testdata.create_file("errpipe.sh", [
            'echo "1" 1>&2',
            '(>&2 echo "2")',
            'echo "3" >&2'
        ])
        os.chmod(filepath, 0o755)
        #cmd = "echo 'foobar' 1>&2"
        #cmd = "/vagrant/errpipe.sh"
        output = cli.run(filepath)
        for x in range(1, 4):
            self.assertTrue(str(x) in output)

    def test_cached_run(self):
        cmd = "date; sleep 1; date"
        r = cli.cached_run(cmd)

        start = time.time()
        r2 = cli.cached_run(cmd)
        stop = time.time()
        self.assertEqual(r, r2)
        self.assertTrue((stop - start) < 1)

        time.sleep(1.1)
        r3 = cli.cached_run(cmd, ttl=1)
        self.assertNotEqual(r, r3)

    def test_ip(self):
        ip = cli.ip()
        ip2 = cli.ip()



# thanks to
# http://seasonofcode.com/posts/setting-up-dkim-and-srs-in-postfix.html
# http://serverfault.com/questions/82234/srs-sender-rewriting-when-forwarding-mail-through-postfix
# http://www.mind-it.info/2014/02/22/forward-postfix-spf-srs/
# http://www.openspf.org/SRS

import re

from .. import cli
from ..path import Dirpath, Filepath, Sentinal
from .base import Interface


class SRS(Interface):
    @property
    def temp_d(self):
        return Dirpath.get_temp()

    @property
    def build_d(self):
        build_d = Dirpath(self.temp_d, "postsrsd-master", "build")
        return build_d

    def start(self):
        try:
            cli.run("start postsrsd")
        except RuntimeError as e:
            if not re.search("Job\s+is\s+already\s+running", str(e), re.I):
                raise

    def restart(self):
        if self.is_running():
            cli.run("restart postsrsd")

        else:
            self.start()

    def stop(self):
        try:
            cli.run("stop postsrsd")
        except cli.RunError as e:
            if not e.search("Unknown\s+instance"):
                raise

    def is_running(self):
        ret = True
        try:
            output = cli.run("status postsrsd")
            if re.search("stop", output, flags=re.I):
                ret = False

        except RuntimeError:
            ret = False

        return ret

    def exists(self):
        ret = True
        try:
            cli.run("status postsrsd")
        except RuntimeError:
            ret = False

        return ret

    def install(self):
        #cli.package("unzip", "cmake", "curl", "build-essential")
        cli.package("unzip", "cmake", "build-essential")

        tmpdir = self.temp_d

        # only download the srs if it doesn't already exist
        postsrs_f = Filepath(tmpdir, "postsrsd.zip")
        with Sentinal.check("srs") as s:
            if not s:
                # https://github.com/roehling/postsrsd
                cli.run("wget -O postsrsd.zip https://github.com/roehling/postsrsd/archive/master.zip", cwd=tmpdir.path)
                #cli.run("curl -L -o postsrsd.zip https://github.com/roehling/postsrsd/archive/master.zip", cwd=tmpdir.path)

        # Build and install.
        cli.run("unzip -o postsrsd.zip", cwd=tmpdir.path)
        build_d = self.build_d
        build_d.create()

        cli.run("cmake -DCMAKE_INSTALL_PREFIX=/usr ../", cwd=build_d.path)
        cli.run("make", cwd=build_d.path)
        cli.run("make install", cwd=build_d.path)

    def uninstall(self):
        self.stop()





from unittest import TestCase

import testdata

from stockton.geo import IP
from stockton.path import Filepath


class IPTest(TestCase):
    def test_lifecycle(self):
        #raise SkipTest("This test is more for manual testing")
        ip = IP()
        self.assertTrue(ip.state)

        db = Filepath(ip.geo_path)
        self.assertTrue(db.exists())

        #pout.v(ip.state)
        #pout.v(ip.country)
        #pout.v(ip.city)






from .. import cli
from ..path import Filepath
from ..concur.formats.postfix import SMTPd


class SMTP(object):
    @property
    def db_f(self):
        return Filepath("/etc/sasldb2")

    @property
    def config_f(self):
        return Filepath(SMTPd.dest_path)

    @property
    def config(self):
        config_f = self.config_f
        if config_f.exists():
            s = SMTPd(prototype_path=config_f.path)
        else:
            s = SMTPd()
        return s

    def add_user(self, username, password, domain):
        """adds an smtp user with credentials:

            user: username@domain
            pass: password

        username -- string -- the smtp username
        password -- string -- the smtp password
        domain -- string -- the domain the username belongs to
        """
        cli.run("echo \"{}\" | saslpasswd2 -c -u {} {} -p".format(password, domain, username))

        sasldb2 = self.db_f
        sasldb2.chmod(400)
        sasldb2.chown("postfix")

    def exists(self):
        return self.db_f.exists()

    def install(self):
        cli.package("sasl2-bin", "libsasl2-modules")

    def uninstall(self):
        cli.purge("sasl2-bin", "libsasl2-modules")
        self.config_f.delete()
        self.db_f.delete()




"""
DNS fun stuff

some relevant reading:
http://stackoverflow.com/questions/19322962/how-can-i-list-all-dns-records
"""
import re
import subprocess
import tempfile
import urllib
import gzip
import os

from .interface.dkim import DKIM
from . import cli


class Record(object):
    pass


class Mailserver(Record):

    def __init__(self, domain, ip):
        self.domain = Domain(str(domain))
        self.ip = Domain(ip)

    def needed_a(self):
        ret = []
        a = self.domain.a(self.ip.host)
        if not a:
            ret = [
                ("hostname", self.domain.host),
                ("target", self.ip.host),
            ]
        return ret

    def needed_ptr(self):
        ret = []
        ptr = self.ip.ptr(self.domain.host)
        if not ptr:
            ret = [
                ("hostname", self.domain.host),
            ]
        return ret


class Alias(Record):
    def __init__(self, domain, mailserver):
        self.domain = Domain(str(domain))
        self.mailserver = Domain(str(mailserver))

    def needed_mx(self):
        ret = []
        mx = self.domain.mx(self.mailserver.host)
        if not mx:
            ret = [
                ("hostname", self.domain.host),
                ("mailserver domain", self.mailserver.host),
                ("priority", "100"),
            ]
        return ret

    def needed_spf(self):
        ret = []
        txts = self.domain.spf()
        if not txts:
            ret = [
                ("hostname", self.domain.host),
                ("text", "v=spf1 mx ~all"),
            ]
        return ret

    def needed_dkim(self):
        ret = []
        dk = DKIM()
        domainkey = dk.domainkey(self.domain.host)
        d = Domain(domainkey.subdomain)
        txts = d.dkim()
        if not txts or (domainkey.p not in txts[0]["text"]):
            ret = [
                ("hostname", domainkey.subdomain),
                ("text", domainkey.text),
            ]
        return ret


class Domain(object):
    def __init__(self, host):
        self.host = host

    def records(self, regex, output, filter_regex=""):
        filter_regexp = re.compile(filter_regex, re.I) if filter_regex else None
        regexp = re.compile(regex, re.I)
        records = filter(None, output.splitlines(False))

        for record in records:
            check_record = record
            if filter_regexp:
                check_record = record if filter_regexp.search(record) else ""

            if check_record:
                m = regexp.search(check_record)
                if m:
                    yield m.groups()

    def nameservers(self, filter_regex=""):
        output = self.query("NS")
        records = [m[0] for m in self.records(
            "^{}\s+name\s+server\s+(.+?)\.?$".format(self.host),
            output,
            filter_regex
        )]
        return records

    def mx(self, filter_regex=""):
        ret = []
        output = self.query("mx")
        records = self.records(
            "^{}\s+mail\s+is\s+handled\s+by\s+(\d+)\s+(.+?)\.?$".format(self.host),
            output,
            filter_regex
        )
        for m in records:
            ret.append({
                "hostname": self.host,
                "priority": m[0],
                "target": m[1]
            })

        return ret

    def a(self, filter_regex=""):
        output = self.query("a")
        records = [m[0] for m in self.records(
            "^{}\s+has\s+address\s+((?:\d+\.){{3}}\d+)$".format(self.host),
            output,
            filter_regex
        )]
        return records

    def txt(self, filter_regex=""):
        ret = []
        output = self.query("txt")
        records = self.records(
            "^{}\s+descriptive\s+text\s+\"(.*?)\"$".format(self.host),
            output,
            filter_regex
        )
        for m in records:
            ret.append({
                "hostname": self.host,
                "text": m[0],
            })

        return ret

    def dkim(self):
        txts = self.txt("=dkim")
        if txts:
            for i in range(len(txts)):
                txts[i]["text"] = re.sub("\"\s+\"", "", txts[i]["text"])
                txts[i]["text"] = re.sub("\\\\;", ";", txts[i]["text"])
        return txts

    def spf(self):
        txts = self.txt("=spf")
        return txts

    def ptr(self, filter_regex=""):
        output = self.query("PTR")
        records = [m[0] for m in self.records(
            "^.*?\s+domain\s+name\s+pointer\s+(.+?)\.?$",
            output,
            filter_regex
        )]
        return records

    def rdns(self):
        ret = []
        ips = self.a()
        for ip in ips:
            d = type(self)(ip)
            for ptr in d.ptr():
                if self.host in ptr:
                    ret.append(ip)
        return ret

    def query(self, record):
        output = ""
        try:
            output = subprocess.check_output([
                "host",
                #"-t {}".format(record),
                "-t",
                record,
                self.host
            ])
        except subprocess.CalledProcessError:
            pass

        return output

#     def records(self, record):
#         output = self.query(record)
#         return filter(None, output.splitlines(False))




"""
Generic starter config parsers for a variety of different file formats

https://en.wikipedia.org/wiki/Configuration_file
http://stackoverflow.com/questions/1925305/best-config-file-format

"""
import re

import base


class EqualOption(base.ConfigOption):
    def format_set(self, name, divider, val):
        option_set_format = "{} {} {}" # key = val
        s = option_set_format.format(name, divider, val)
        return s


class EqualConfig(base.Config):
    """Configuration files that have a format like

    name = val

    type of configuration
    """
    option_divider = "="
    option_class = EqualOption


class ColonOption(base.ConfigOption):
    def format_set(self, name, divider, val):
        option_set_format = "{}{} {}" # key: val
        s = option_set_format.format(name, divider, val)
        return s


class ColonConfig(base.Config):
    """Configuration files that have a format like

    name: val

    type of configuration
    """
    option_divider = ":"
    option_class = ColonOption


class SpaceOption(base.ConfigOption):
    def format_set(self, name, divider, val):
        option_set_format = "{}{}" # key val
        s = option_set_format.format(
            name.ljust(max(self.config.option_buffer, len(name) + 1), divider),
            val
        )
        return s

    def _parse(self, fp):
        line = fp.line
        name = ""
        val = ""
        comment = ""
        commenters = self.config.commenters
        divider = self.config.option_divider

        if re.match("^[^{}]\S+".format(commenters), line): # name val
            name, val = re.split("\s+", line, 1)

        elif re.match("^[{}]\s*\S+\s*{}".format(commenters, divider), line): # # name val
            l = re.sub("^[{}]\s*".format(commenters), "", line)
            name, val = re.split("\s+", l, 1)

        #elif re.match("^[{}]".format(commenters)):

        bits = re.split("\s[{}]\s*".format(commenters), val, 1)
        val = bits[0]
        if len(bits) > 1:
            comment = bits[1]

        self.name = name
        self.val = val
        self.comment = comment
        self.line = line
        self.modified = False


class SpaceConfig(base.Config):
    """Configuration files that have a format like

    name            val

    type of configuration
    """
    option_divider = " "
    option_buffer = 24
    option_class = SpaceOption




#!/usr/bin/env python
# http://docs.python.org/distutils/setupscript.html
# http://docs.python.org/2/distutils/examples.html

from setuptools import setup, find_packages
import re
import os


name = "stockton"
with open(os.path.join(name, "__init__.py")) as f:
    version = re.search("^__version__\s*=\s*[\'\"]([^\'\"]+)", f.read(), flags=re.I | re.M).group(1)

setup(
    name=name,
    version=version,
    description='Easy peasy email forwarding for a custom domain',
    author='Jay Marcyes',
    author_email='jay@marcyes.com',
    url='http://github.com/jaymon/{}'.format(name),
    packages=find_packages(),
    license='GPLv2+',
    install_requires=['geoip2', 'captain'],
    tests_require=['testdata'],
    classifiers=[ # https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Communications :: Email',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)'
    ],
    entry_points = {
        'console_scripts': [
            '{} = {}.__main__:console'.format(name, name),
        ],
    },
)



import re

from captain import echo

from .. import cli
from .postfix import Postfix
from ..path import Filepath, Dirpath
#from ..concur.formats.opendkim import OpenDKIM
from ..concur.formats.spamassassin import SpamAssassin, Local


class Spam(object):

    @property
    def user(self):
        return "debian-spamd"

    @property
    def home_d(self):
        # http://askubuntu.com/questions/410244/a-command-to-list-all-users-and-how-to-add-delete-modify-users
        username = self.user
        path = cli.run(
            'grep "{}" /etc/passwd | cut -f6 -d":"'.format(username),
            capture_output=True
        ).strip()

        if not path:
            raise ValueError("could not find home directory for user {}".format(username))

        return Dirpath(path)

    @property
    def config_f(self):
        return Filepath(SpamAssassin.dest_path)

    @property
    def local_f(self):
        return Filepath(Local.dest_path)

    def base_configs(self):
        return [self.config_f, self.local_f]

    def config(self, path=SpamAssassin.dest_path):
        return SpamAssassin(prototype_path=path)

    def local(self, path=Local.dest_path):
        return Local(prototype_path=path)

    def start(self):
        try:
            o = cli.run("/etc/init.d/spamassassin start")
        except RuntimeError as e:
            if not re.search("already\s+running", str(e), re.I):
                raise

    def restart(self):
        if self.is_running():
            cli.run("/etc/init.d/spamassassin restart")
        else:
            self.start()

    def stop(self):
        cli.run("/etc/init.d/spamassassin stop")

    def is_running(self):
        ret = True
        try:
            cli.run("/etc/init.d/spamassassin status")
        except RuntimeError:
            ret = False
        return ret

    def exists(self):
        ret = True
        try:
            self.stop()
        except cli.RunError as e:
            ret = not e.is_missing()
        return ret


    def install(self):
        # make sure the packages are installed
        cli.package("spamassassin", "spamc")

        # add the spam user
        # update 4-16-2016 let's just use debian-spamd and /var/lib/spamassassin
        #cli.run("adduser --disabled-login --disabled-password --gecos "" spamd")
        if not self.home_d.exists():
            raise ValueError("Home directory {} does not exist".format(self.home_d))

    def uninstall(self):
        try:
            self.stop()
        except cli.RunError as e:
            if not e.is_missing():
                raise

        cli.purge("spamassassin", "spamc")

